
import itertools
import numpy

import iotbx.pdb
import iotbx.pdb.amino_acid_codes
import mmtbx.alignment

from scitbx.math import superpose
from scitbx.array_family import flex

from giant.structure import make_label
from giant.structure.select import protein

def nearby_coords_bool(query, coords, cutoff):
    """Find all points in coords within cutoff of query. Return boolean selection"""
    assert isinstance(coords, flex.vec3_double)
    return (coords-query).dot() < (cutoff**2)

def trim_structures_to_common_atoms(hierarchy_1, hierarchy_2):
    """Reduce a pair of structures to the common set of atoms - only protein residues and the first conformer are taken"""

    # List of atoms to copy from hierarchy_1 and hierarchy_2
    common_set_1 = []
    common_set_2 = []
    # Iterate through residue_groups and find residues that are present in both hierarchies
    for ch_1 in hierarchy_1.chains():
        for rg_1 in ch_1.residue_groups():
            conf_1 = rg_1.conformers()[0]
    return hierarchy_1, hierarchy_2

def perform_flexible_alignment(mov_hierarchy, ref_hierarchy, cutoff_radius=10, require_non_protein_identical=True):
    """Perform a flexible alignment on the two hierarchies. Will return a list of alignment metrics, one for each residue."""

    try:
        mov_hierarchy.only_model()
        ref_hierarchy.only_model()
    except:
        raise Exception('structures for alignment can only have one model!')

    # Filter the structures to only the protein (don't care about waters)
    if not require_non_protein_identical:
        # Extract caches
        ref_cache = ref_hierarchy.atom_selection_cache()
        mov_cache = mov_hierarchy.atom_selection_cache()
        # Strip structures to protein only
        ref_hierarchy = ref_hierarchy.select(ref_cache.iselection('pepnames'))
        mov_hierarchy = mov_hierarchy.select(mov_cache.iselection('pepnames'))

    # Check the structures are identical
    assert mov_hierarchy.is_similar_hierarchy(ref_hierarchy), 'HIERARCHIES MUST BE SIMILAR'

    # Extract caches
    ref_cache = ref_hierarchy.atom_selection_cache()
    mov_cache = mov_hierarchy.atom_selection_cache()

    # Caches for the reference structure
    ref_backbone_hier  = ref_hierarchy.select(ref_cache.iselection('pepnames and (name CA or name C or name O or name N)'))
    ref_backbone_cache = ref_backbone_hier.atom_selection_cache()
    # Caches for the query structure
    mov_backbone_hier  = mov_hierarchy.select(mov_cache.iselection('pepnames and (name CA or name C or name O or name N)'))
    mov_backbone_cache = mov_backbone_hier.atom_selection_cache()

    # Tranformation dictionaries
    rts = {}

    # Calculate a transformation matrix for each calpha
    for calpha_at in ref_hierarchy.select(ref_cache.iselection('pepnames and name CA')).atoms_with_labels():
        # Get selection within cutoff distance
        align_sel = nearby_coords_bool( query  = calpha_at.xyz,
                                        coords = ref_backbone_hier.atoms().extract_xyz(),
                                        cutoff = cutoff_radius  )
        # Extract full atom groups where one of the atoms has been selected
        align_ags = ref_backbone_hier.select(align_sel).atom_groups()
        align_str = ' or '.join(['(chain "{}" and resid "{}" and altloc "{}")'.format(ag.parent().parent().id, ag.parent().resid(), ag.altloc if ag.altloc else ' ') for ag in align_ags])
        # Extract atoms to be aligned
        ref_align_ats = ref_backbone_hier.select(ref_backbone_cache.selection(align_str)).atoms()
        mov_align_ats = mov_backbone_hier.select(mov_backbone_cache.selection(align_str)).atoms()
        # Check that atoms are identical in the two hierarchies
        assert len(ref_align_ats) == len(mov_align_ats), 'atoms to be aligned are not the same length'
        assert len(ref_align_ats) > 0, len(ref_align_ats)
        assert len(ref_align_ats) >= sum(align_sel), 'Aligned atoms fewer than select atoms: {} < {}'.format(len(ref_align_ats), sum(align_sel))
        for ref_at, mov_at in zip(ref_align_ats, mov_align_ats):
            assert ref_at.pdb_label_columns() == mov_at.pdb_label_columns(), 'ATOMS MUST BE IDENTICAL'
        # Calculate the alignment for this residue
        rt = superpose.least_squares_fit(   reference_sites = ref_align_ats.extract_xyz(),
                                            other_sites     = mov_align_ats.extract_xyz()   ).rt()
        # Save the rotation matrix
        rts[make_label(calpha_at)] = rt
    return rts

def find_nearest_calphas(hierarchy, coordinates):
    """Find the nearest calpha in hierarchy for each coordinate in coordinates"""
    import scipy.spatial
    # Extract calphas and associated labels
    hierarchy = hierarchy.select(hierarchy.atom_selection_cache().selection('pepnames and name CA'))
    calpha_sites, calpha_labs = zip(*[(a.xyz, make_label(a)) for a in hierarchy.atoms_with_labels()])
    # Build a tree to find the smallest distance
    tree = scipy.spatial.KDTree(data=calpha_sites)
    # Find the nearest neighbours
    nn_dists, nn_groups = tree.query(coordinates)
    # Extract the associated labels
    mappings = [calpha_labs[i] for i in nn_groups]
    return mappings

def transform_coordinates_with_flexible_alignment(alignments, coordinates, mappings, inverse=False):
    """Transform coordinates by associated alignments associated with mappings values"""

    assert len(coordinates) == len(mappings)
    if not isinstance(coordinates, flex.vec3_double):
        coordinates = flex.vec3_double(coordinates)
    # Sort the indices by the mapping values
    num_tot = len(coordinates)
    sorted_idxs     = flex.size_t(sorted(range(num_tot), key=lambda i: mappings[i]))
    sorted_coords   = coordinates.select(sorted_idxs)
    sorted_mappings = [mappings[i] for i in sorted_idxs]

    # Initialise output array
    out_coords = numpy.zeros(len(coordinates), dtype=[('x',float),('y',float),('z',float)])

    # Iterate through the coords in groups and transform
    for lab, lab_idxs in itertools.groupby(range(num_tot), key=lambda i: sorted_mappings[i]):
        lab_idxs = flex.size_t(lab_idxs)
#        print 'Using RT for {}'.format(lab)
#        print 'on {} points'.format(len(lab_idxs))
#        print 'from idx {} to {}'.format(lab_idxs[0], lab_idxs[-1])
        # Extract coordinates for this block
        lab_coords   = sorted_coords.select(lab_idxs)
        lab_mappings = [sorted_mappings[i] for i in lab_idxs]
        orig_idxs    = sorted_idxs.select(lab_idxs)
        assert max(lab_idxs)-min(lab_idxs) == len(lab_idxs)-1
        assert len(set(lab_mappings)) == 1
        # Extract RT matrix
        rt = alignments[lab]
        if inverse: rt = rt.inverse()
        # Transform and put back
        rt_coords = rt * lab_coords
        out_coords.put(orig_idxs, rt_coords)

    return flex.vec3_double(out_coords)

def extract_sites_for_alignment(chain_obj):
    """Extract sequence and sites of c-alphas - adapted from mmtbx.command_line.super"""

    seq = []
    sites = flex.vec3_double()
    use_sites = flex.bool()
    for resi in chain_obj.conformers()[0].residues():
        if (   iotbx.pdb.common_residue_names_get_class(name=resi.resname) != "common_amino_acid"):
            continue
        resn = resi.resname
        single = iotbx.pdb.amino_acid_codes.one_letter_given_three_letter[resn]
        seq.append(single)
        use = False
        xyz = (0,0,0)
        for atom in resi.atoms():
            if (atom.name == " CA "):
              xyz = atom.xyz
              use = True
              break
        sites.append(xyz)
        use_sites.append(use)
    return "".join(seq), sites, use_sites

def align_hierarchies(mov_hier, ref_hier):
    """Extract c-alpha sites from the structures and align"""
    return align_chains(mov_chain=protein(mov_hier, copy=True).models()[0].only_chain(),
                        ref_chain=protein(ref_hier, copy=True).models()[0].only_chain() )

def align_chains(mov_chain, ref_chain):
    """Takes two chains and aligns them - return rt_mx"""

    mov_seq, mov_sites, mov_flags = extract_sites_for_alignment(mov_chain)
    ref_seq, ref_sites, ref_flags = extract_sites_for_alignment(ref_chain)

    align_obj = mmtbx.alignment.align(
        seq_a=ref_seq,
        seq_b=mov_seq,
        gap_opening_penalty = 20,
        gap_extension_penalty = 2,
        similarity_function = 'blosum50',
        style = 'local')

    # Extract the alignment
    alignment = align_obj.extract_alignment()
    # List of matches - '|' for exact match, '*' for good match
    matches = alignment.matches()
    equal = matches.count("|")
    similar = matches.count("*")
    total = len(alignment.a) - alignment.a.count("-")
    alignment.pretty_print(
        matches=matches,
        block_size=50,
        n_block=1,
        top_name="fixed",
        bottom_name="moving")

    # Create list of selected sites
    ref_sites_sel = flex.vec3_double()
    mov_sites_sel = flex.vec3_double()
    for ia,ib,m in zip(alignment.i_seqs_a, alignment.i_seqs_b, matches):
        if (m not in ["|", "*"]): continue
        # Check that the sites are flagged to be used
        if (ref_flags[ia] and mov_flags[ib]):
            # Append sites to list to align
            ref_sites_sel.append(ref_sites[ia])
            mov_sites_sel.append(mov_sites[ib])

    if (ref_sites_sel.size() == 0):
      raise Exception("No matching C-alpha atoms.")

    # Create LSQ
    lsq_fit = superpose.least_squares_fit(
        reference_sites = ref_sites_sel,
        other_sites     = mov_sites_sel)

    return lsq_fit



import numpy
from scitbx.array_family import flex
from libtbx.math_utils import ifloor, iceil

def get_bounding_box_for_structure(structure):
    return (structure.sites_cart().min(), structure.sites_cart().max())

def calculate_grid_size(min_carts, max_carts, grid_spacing):
    """Calculate the number of points to be sampled for a box size and sampling distance. Returns the size of the box in A, and the number of points to be sampled along each axis."""

    # Extent of the box (dX, dY, dZ)
    cart_size = tuple([max_c-min_c for min_c, max_c in zip(min_carts, max_carts)])

    # Calculate the number of grid points from the grid sampling
    grid_size = tuple([iceil((1.0*c_size)/grid_spacing)+1 for c_size in cart_size])

    return cart_size, grid_size

def create_cartesian_grid(min_carts, max_carts, grid_spacing):
    """Create the grid of cartesian sites to be sampled based on a spacing between sampling points (in A)"""

    # Calculate the size of the grid (cartesian) and the number of points to be sampled along each axis
    box_size, grid_size = calculate_grid_size(min_carts, max_carts, grid_spacing)

    # Grid point volume
    grid_point_volume = grid_spacing**3

    # Calculate the grid points in cartesian space
    grid_points_cart = flex.vec3_double(flex.nested_loop(grid_size)) * grid_spacing + min_carts
#    grid_points_cart = flex.vec3_double([tuple([min_carts[i]+grid_spacing*grid_point[i] for i in [0,1,2]]) for grid_point in flex.nested_loop(grid_size)])

    return box_size, grid_size, grid_points_cart

def get_grid_points_within_distance_cutoff_of_origin(grid_spacing, distance_cutoff):
    """Find all points on isotropic grid within distance_cutoff of the origin"""

    # Normalise the cutoff to make the grid spacing grid-independent
    grid_index_cutoff = (1.0*distance_cutoff)/grid_spacing

    return get_grid_points_within_index_cutoff_of_origin(grid_index_cutoff=grid_index_cutoff)

def get_grid_points_within_index_cutoff_of_origin(grid_index_cutoff):
    """Find all points relative to the origin within a certain number of grid points"""

    # Round the max grid index down to the nearest int - outer bound on the x,y,z coords
    outer_bound_box = int(grid_index_cutoff)
    # Calculate r/sqrt(3) - inner bound on x,y,z coords
    inner_bound_box = grid_index_cutoff/numpy.math.sqrt(3)
    # Calculate r^2 - limiting sphere
    rad_sq = grid_index_cutoff**2
    # List of allowed grid indices
    grid_points = []

    for x,y,z in flex.nested_loop(begin=(-outer_bound_box, -outer_bound_box, -outer_bound_box),end=(outer_bound_box+1, outer_bound_box+1, outer_bound_box+1)):
        if (abs(x) <= inner_bound_box) and (abs(y) <= inner_bound_box) and (abs(z) <= inner_bound_box):
            grid_points.append((x, y, z))
        elif (x**2 + y**2 + z**2) <= rad_sq:
            grid_points.append((x, y, z))

    return grid_points

def combine_grid_point_and_grid_vectors(start_point, grid_vectors):
    """Take a list of grid vectors and add them to a particular grid point"""

    start_point = flex.int(start_point)
    grid_points = [tuple(start_point+flex.int(vec)) for vec in grid_vectors]
    return grid_points

def get_grid_points_within_distance_cutoff_of_cart_sites(cart_sites, grid_spacing, max_dist, min_dist=None):
    """Find all points on isotropic grid within distance cutoff of cartesian sites"""

    # Normalise the cutoff and the sites to make them grid-independent
    max_grid_dist = (1.0*max_dist)/grid_spacing
    if min_dist: min_grid_dist = (1.0*min_dist)/grid_spacing
    else:        min_grid_dist = None
    grid_sites = [tuple([1.0*c/grid_spacing for c in coords]) for coords in cart_sites]

    return get_grid_points_within_index_cutoff_of_grid_sites(grid_sites=grid_sites, max_grid_dist=max_grid_dist, min_grid_dist=min_grid_dist)

def get_grid_points_within_distance_cutoff_of_cart_sites_2(cart_sites, grid_spacing, max_dist, min_dist=None):
    """Find all points on isotropic grid within distance cutoff of cartesian sites"""

    # Normalise the cutoff and the sites to make them grid-independent
    max_grid_dist = (1.0*max_dist)/grid_spacing
    if min_dist: min_grid_dist = (1.0*min_dist)/grid_spacing
    else:        min_grid_dist = None
    grid_sites = [tuple([1.0*c/grid_spacing for c in coords]) for coords in cart_sites]

    return get_grid_points_within_index_cutoff_of_grid_sites_2(grid_sites=grid_sites, max_grid_dist=max_grid_dist, min_grid_dist=min_grid_dist)

def get_grid_points_within_index_cutoff_of_grid_sites(grid_sites, max_grid_dist, min_grid_dist=None):
    """Find all points on a grid within a certain number of grid points of grid sites (not necessarily integer sites)"""

    # Calculate the size of the grid we need to check over
    min_x = ifloor(min([s[0] for s in grid_sites]) - max_grid_dist)
    if min_x < 0: min_x=0
    min_y = ifloor(min([s[1] for s in grid_sites]) - max_grid_dist)
    if min_y < 0: min_y=0
    min_z = ifloor(min([s[2] for s in grid_sites]) - max_grid_dist)
    if min_z < 0: min_z=0
    max_x = iceil(max([s[0] for s in grid_sites]) + max_grid_dist)
    max_y = iceil(max([s[1] for s in grid_sites]) + max_grid_dist)
    max_z = iceil(max([s[2] for s in grid_sites]) + max_grid_dist)
    # Grid Extremities
    min_grid = (min_x, min_y, min_z)
    max_grid = (max_x+1, max_y+1, max_z+1)

    # Round the max grid distance down to the nearest int - outer bound on the dx,dy,dz values
    outer_bound_box = int(max_grid_dist)
    # Calculate r/sqrt(3) - inner bound on dx, dy, dz values
    inner_bound_box = max_grid_dist/numpy.math.sqrt(3)
    # Calculate r^2 - limiting sphere
    rad_sq = max_grid_dist**2

    # List of allowed grid indices
    outer_points = []
    inner_points = []

    # Iterate through and add valid points
    for gp in flex.nested_loop(min_grid, max_grid):
        for site in grid_sites:
            dx, dy, dz = [abs(p1-p2) for p1,p2 in zip(gp, site)]

            if (dx > outer_bound_box) or (dy > outer_bound_box) or (dz > outer_bound_box):
                continue
            elif (dx <= inner_bound_box) and (dy <= inner_bound_box) and (dz <= inner_bound_box):
                outer_points.append(gp)
                break
            elif (dx**2 + dy**2 + dz**2) <= rad_sq:
                outer_points.append(gp)
                break

    # Filter the grid points that are too close to the protein
    if min_grid_dist:
        # Round the min grid distance up to the nearest int - outer bound on the dx,dy,dz values
        outer_bound_box = int(min_grid_dist) + 1
        # Calculate r/sqrt(3) - inner bound on dx, dy, dz values
        inner_bound_box = min_grid_dist/numpy.math.sqrt(3)
        # Calculate r^2 - limiting sphere
        rad_sq = min_grid_dist**2

        # Iterate through and add valid points
        for gp in outer_points:
            for site in grid_sites:
                dx, dy, dz = [abs(p1-p2) for p1,p2 in zip(gp, site)]

                if (dx > outer_bound_box) or (dy > outer_bound_box) or (dz > outer_bound_box):
                    continue
                elif (dx <= inner_bound_box) and (dy <= inner_bound_box) and (dz <= inner_bound_box):
                    inner_points.append(gp)
                    break
                elif (dx**2 + dy**2 + dz**2) <= rad_sq:
                    inner_points.append(gp)
                    break

    if min_grid_dist:
        total_point = [gp for gp in outer_points if gp not in inner_points]
    else:
        total_points = outer_points

    return total_points, outer_points, inner_points

def get_grid_points_within_index_cutoff_of_grid_sites_2(grid_sites, max_grid_dist, min_grid_dist=None):
    """Find all points on a grid within a certain number of grid points of grid sites (not necessarily integer sites)"""

    # Find the size of the grid that we'll need
    max_x = iceil(max([s[0] for s in grid_sites]) + max_grid_dist)
    max_y = iceil(max([s[1] for s in grid_sites]) + max_grid_dist)
    max_z = iceil(max([s[2] for s in grid_sites]) + max_grid_dist)
    max_grid = (max_x+2, max_y+2, max_z+2)

    # Round the grid sites to the nearest grid point
    int_grid_sites = [tuple([int(round(x,0)) for x in site]) for site in grid_sites]

    # Grid objects
    grid_indexer = flex.grid(max_grid)
    grid_size = flex.product(flex.int(max_grid))

    # Outer mask
    outer_indices_mask = numpy.zeros(grid_size, dtype=int)

    # Find all of the grid vectors within max_dist of grid sites
    outer_grid_vectors = get_grid_points_within_index_cutoff_of_origin(grid_index_cutoff=max_grid_dist)
    outer_grid_points = []
    for site in int_grid_sites:
        outer_grid_points.extend(combine_grid_point_and_grid_vectors(site, grid_vectors=outer_grid_vectors))

    # They may overlap, so find unique grid_points
    outer_grid_points = list(set(outer_grid_points))
    # Filter those that are outside of the grid
    outer_grid_points = [gp for gp in outer_grid_points if not ( ([i+1 for i in range(3) if gp[i]<0]) or ([i+1 for i in range(3) if gp[i]>max_grid[i]]) )]
    # Map the grid points to the associated 1d index
    outer_grid_indices = [grid_indexer(gp) for gp in outer_grid_points]
    # Create a binary mask of the points
    [outer_indices_mask.put(i, 1) for i in outer_grid_indices]

    # Inner mask
    inner_indices_mask = numpy.zeros(grid_size, dtype=int)
    if min_grid_dist:
        # Find all of the grid vectors within min_dist of grid sites
        inner_grid_vectors = get_grid_points_within_index_cutoff_of_origin(grid_index_cutoff=min_grid_dist)
        inner_grid_points = []
        for site in int_grid_sites:
            inner_grid_points.extend(combine_grid_point_and_grid_vectors(site, grid_vectors=inner_grid_vectors))

        # They may overlap, so find unique grid_points
        inner_grid_points = list(set(inner_grid_points))
        # Filter those that are outside of the grid
        inner_grid_points = [gp for gp in inner_grid_points if not ( ([i+1 for i in range(3) if gp[i]<0]) or ([i+1 for i in range(3) if gp[i]>=max_grid[i]]) )]
        # Map the grid points to the associated 1d index
        inner_grid_indices = [grid_indexer(gp) for gp in inner_grid_points]
        # Create a binary mask of the points
        [inner_indices_mask.put(i, 1) for i in inner_grid_indices]

    # Convert from the mask back to grid points
    outer_points = [gp for i, gp in enumerate(flex.nested_loop(max_grid)) if outer_indices_mask[i]==1]
    inner_points = [gp for i, gp in enumerate(flex.nested_loop(max_grid)) if inner_indices_mask[i]==1]
    total_points = [gp for i, gp in enumerate(flex.nested_loop(max_grid)) if (inner_indices_mask[i]==0 and outer_indices_mask[i]==1)]

    return total_points, outer_points, inner_points






import os, sys

#################################
try:
    import matplotlib
    matplotlib.use('Agg')
    matplotlib.interactive(0)
    from matplotlib import pyplot
    pyplot.style.use('ggplot')
except:
    pass
#################################

import numpy

from scitbx.array_family import flex

def filter_nans(x):
    return [v for v in x if not numpy.isnan(v)]

def map_value_distribution(f_name, plot_vals, plot_normal):
    """Plot histogram of values, with optional normal distribution"""
    from scitbx.math.distributions import normal_distribution
    fig = pyplot.figure()
    pyplot.title('Map Value Distribution')
    pyplot.hist(x=plot_vals, bins=30, normed=True)
    if plot_normal:
        # Plot the distribution for N(0,1)
        nd_t = normal_distribution()
        theor_x = numpy.linspace(-5,5,101)
        theor_y = [nd_t.pdf(x) for x in theor_x]
        pyplot.plot(theor_x, theor_y, c='k', ls='--', marker='o')
        # Plot the distribution for the observed distribution
        nd_o = normal_distribution(mean=numpy.mean(plot_vals), sd=numpy.std(plot_vals))
        obs_x = numpy.linspace(-5,5,101)
        obs_y = [nd_o.pdf(x) for x in obs_x]
        pyplot.plot(obs_x, obs_y, c='g', ls='-', marker='o')
    pyplot.xlabel('Map Value')
    pyplot.ylabel('Density')
    pyplot.tight_layout()
    pyplot.savefig(f_name)
    pyplot.close(fig)

def qq_plot_against_normal(f_name, plot_vals):
    """Sort and plot list of values against expected quantiles from a normal distribution"""
    from scitbx.math.distributions import normal_distribution
    fig = pyplot.figure()
    pyplot.title('Observed v Theoretical Q-Q Plot')
    expected_vals = normal_distribution().quantiles(len(plot_vals))
    pyplot.plot([min(expected_vals)-1, max(expected_vals)+1], [min(expected_vals)-1, max(expected_vals)+1], 'b--')
    pyplot.plot(sorted(plot_vals), expected_vals, 'go-')
    pyplot.xlabel('Observed Quantiles')
    pyplot.ylabel('Theoretical Quantiles')
    pyplot.tight_layout()
    pyplot.savefig(f_name)
    pyplot.close(fig)

def mean_obs_scatter(f_name, mean_vals, obs_vals):
    """Plot mean map against observed map values"""
    fig = pyplot.figure()
    pyplot.title('Unsorted Mean v Dataset Scatter Plot')
    pyplot.plot([-3, 10], [-3, 10], 'b--')
    pyplot.plot(mean_vals, obs_vals, 'go')
    pyplot.xlabel('Mean Map Values')
    pyplot.ylabel('Dataset Map Values')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(f_name)
    pyplot.close(fig)

def sorted_mean_obs_scatter(f_name, mean_vals, obs_vals):
    """Plot sorted mean map against sorted observed map values"""
    fig = pyplot.figure()
    pyplot.title('Sorted Mean v Dataset Q-Q Scatter Plot')
    pyplot.plot([-3, 10], [-3, 10], 'b--')
    pyplot.plot(mean_vals, obs_vals, 'go')
    pyplot.xlabel('Sorted Mean Map Values')
    pyplot.ylabel('Sorted Dataset Map Values')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(f_name)
    pyplot.close(fig)

def diff_mean_qqplot(f_name, map_off, map_unc, q_cut, obs_diff, quantile):
    """Plot diff-mean map against normal quantiles, with uncertainty lines"""
    fig = pyplot.figure()
    pyplot.title('Mean-Difference Map Q-Q Plot')
    pyplot.plot([map_off-5*map_unc, map_off+5*map_unc], [-5, 5], 'b--')
    pyplot.plot([-1, 1], [q_cut, q_cut], 'k-.')
    pyplot.plot([-1, 1], [-q_cut, -q_cut], 'k-.')
    pyplot.plot(obs_diff, quantile, 'go-')
    pyplot.xlabel('Difference from Mean Map Quantiles')
    pyplot.ylabel('Theoretical Quantiles')
    pyplot.tight_layout()
    pyplot.savefig(f_name)
    pyplot.close(fig)

def write_dataset_summary_graphs(pandda):
    """Plot dataset summary graphs of resolution, unit cell variation, etc"""

    pandda.log('----------------------------------->>>')
    pandda.log('Generating Summary Graphs')

    # Filter the dataset to non-rejected datasets
    non_rejected_dtags = [dh.tag for dh in pandda.datasets.mask(mask_name='rejected - total', invert=True)]

    d_info = pandda.tables.dataset_info.loc[non_rejected_dtags]
    n_bins = 30

    # ================================================>
    fig = pyplot.figure()
    pyplot.title('Resolution Histograms')
    pyplot.subplot(2, 1, 1)
    pyplot.hist(x=d_info['high_resolution'], bins=n_bins)
    pyplot.xlabel('High Resolution Limit (A)')
    pyplot.ylabel('Count')
    pyplot.subplot(2, 1, 2)
    pyplot.hist(x=d_info['low_resolution'], bins=n_bins)
    pyplot.xlabel('Low Resolution Limit (A)')
    pyplot.ylabel('Count')
    pyplot.tight_layout()
    pyplot.savefig(pandda.output_handler.get_file('d_resolutions'))
    pyplot.close(fig)
    # ================================================>
    try:
        fig = pyplot.figure()
        pyplot.title('R-Factor Histograms')
        pyplot.subplot(2, 1, 1)
        pyplot.hist(x=d_info['r_free'], bins=n_bins)
        pyplot.xlabel('R-Free')
        pyplot.ylabel('Count')
        pyplot.subplot(2, 1, 2)
        pyplot.hist(x=d_info['r_work'], bins=n_bins)
        pyplot.xlabel('R-Work')
        pyplot.ylabel('Count')
        pyplot.tight_layout()
        pyplot.savefig(pandda.output_handler.get_file('d_rfactors'))
        pyplot.close(fig)
    except:
        fig = pyplot.figure()
        pyplot.title('Failed to make R-Factor Histograms')
        pyplot.savefig(pandda.output_handler.get_file('d_rfactors'))
        pyplot.close(fig)
    # ================================================>
    fig = pyplot.figure()
    pyplot.title('RMSDs to Reference Structure Histogram')
    pyplot.hist(x=filter_nans(d_info['rmsd_to_reference']), bins=n_bins)
    pyplot.xlabel('RMSD (A)')
    pyplot.ylabel('Count')
    pyplot.tight_layout()
    pyplot.savefig(pandda.output_handler.get_file('d_global_rmsd_to_ref'))
    pyplot.close(fig)
    # ================================================>
    fig = pyplot.figure()
    pyplot.title('Unit Cell Axis Variation')
    pyplot.subplot(3, 1, 1)
    pyplot.hist(x=d_info['uc_a'], bins=n_bins)
    pyplot.xlabel('A (A)')
    pyplot.subplot(3, 1, 2)
    pyplot.hist(x=d_info['uc_b'], bins=n_bins)
    pyplot.xlabel('B (A)')
    pyplot.subplot(3, 1, 3)
    pyplot.hist(x=d_info['uc_c'], bins=n_bins)
    pyplot.xlabel('C (A)')
    pyplot.tight_layout()
    pyplot.savefig(pandda.output_handler.get_file('d_cell_axes'))
    pyplot.close(fig)
    # ================================================>
    fig = pyplot.figure()
    pyplot.title('Unit Cell Angle Variation')
    pyplot.subplot(3, 1, 1)
    pyplot.hist(x=d_info['uc_alpha'], bins=n_bins)
    pyplot.xlabel('Alpha')
    pyplot.subplot(3, 1, 2)
    pyplot.hist(x=d_info['uc_beta'], bins=n_bins)
    pyplot.xlabel('Beta')
    pyplot.subplot(3, 1, 3)
    pyplot.hist(x=d_info['uc_gamma'], bins=n_bins)
    pyplot.xlabel('Gamma')
    pyplot.tight_layout()
    pyplot.savefig(pandda.output_handler.get_file('d_cell_angles'))
    pyplot.close(fig)
    # ================================================>
    fig = pyplot.figure()
    pyplot.title('Unit Cell Volume Variation')
    pyplot.hist(x=d_info['uc_vol'], bins=n_bins)
    pyplot.xlabel('Volume (A^3)')
    pyplot.ylabel('Count')
    pyplot.tight_layout()
    pyplot.savefig(pandda.output_handler.get_file('d_cell_volumes'))
    pyplot.close(fig)

def write_map_analyser_graphs(pandda, map_analyser, analysis_mask_name):

    # Get the output directory to write the graphs into
    img_out_dir = os.path.join(pandda.output_handler.get_dir('analyses'), '{!s}A Maps'.format(map_analyser.meta.resolution))
    if not os.path.exists(img_out_dir): os.mkdir(img_out_dir)

    # Map Analyser Variables
    map_res = map_analyser.meta.resolution

    ########################################################

    # Statistical Map Values

    # Points to extract from the grid
    grid_mask = pandda.reference_grid().global_mask().outer_mask_indices()

    # Extract map values for plotting
    mean_map_vals = list(map_analyser.statistical_maps.mean_map.select(grid_mask))
    try:    medn_map_vals = list(map_analyser.statistical_maps.medn_map.select(grid_mask))
    except: medn_map_vals = mean_map_vals
    stds_map_vals = list(map_analyser.statistical_maps.stds_map.select(grid_mask))
    sadj_map_vals = list(map_analyser.statistical_maps.sadj_map.select(grid_mask))

    ########################################################

    # Dataset Parameters
    d_info = pandda.tables.dataset_info
    m_info = pandda.tables.dataset_map_info

    # All datasets
    high_res = [d_info['high_resolution'][mh.tag] for mh in map_analyser.dataset_maps.all()]
    low_res =  [d_info['low_resolution'][mh.tag]  for mh in map_analyser.dataset_maps.all()]
    rfree =    [d_info['r_free'][mh.tag]          for mh in map_analyser.dataset_maps.all()]
    rwork =    [d_info['r_work'][mh.tag]          for mh in map_analyser.dataset_maps.all()]

    # All datasets
    map_uncties = [m_info['map_uncertainty'][mh.tag] for mh in map_analyser.dataset_maps.all()]
    # Analysed datasets only
    z_map_mean  = [m_info['z_map_mean'][mh.tag] for mh in map_analyser.dataset_maps.mask(mask_name=analysis_mask_name)]
    z_map_std   = [m_info['z_map_std'][mh.tag]  for mh in map_analyser.dataset_maps.mask(mask_name=analysis_mask_name)]
    z_map_skew  = [m_info['z_map_skew'][mh.tag] for mh in map_analyser.dataset_maps.mask(mask_name=analysis_mask_name)]
    z_map_kurt  = [m_info['z_map_kurt'][mh.tag] for mh in map_analyser.dataset_maps.mask(mask_name=analysis_mask_name)]

    ########################################################

    n_bins = 30

    ########################################################

    pandda.log('=> Writing Statistical Map Distributions')

    ##################################
    # STATISTICAL MAP HISTOGRAMS
    ##################################
    fig = pyplot.figure()
    pyplot.title('Mean and Median Map Histograms')
    # MEAN MAP
    pyplot.subplot(2, 1, 1)
    pyplot.hist(x=mean_map_vals, bins=n_bins)
    pyplot.xlabel('Mean Map Values')
    pyplot.ylabel('Count')
    # MEDIAN MAP
    pyplot.subplot(2, 1, 2)
    pyplot.hist(x=medn_map_vals, bins=n_bins)
    pyplot.xlabel('Median Map Values')
    pyplot.ylabel('Count')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save both
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-mean_medn_map_hist.png'.format(map_res)))
    pyplot.close(fig)

    ##################################
    # MEAN Values v MEDIAN Values
    ##################################
    fig = pyplot.figure()
    pyplot.title('Mean v Median Map Values')
    pyplot.scatter(x=mean_map_vals, y=medn_map_vals)
    # Plot straight line between the min and max values
    min_val = min(mean_map_vals+medn_map_vals)
    max_val = max(mean_map_vals+medn_map_vals)
    pyplot.plot([min_val, max_val], [min_val, max_val], 'b--')
    # Axis labels
    pyplot.xlabel('Mean Map Value')
    pyplot.ylabel('Median Map Value')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-mean_v_median_scatter.png'.format(map_res)))
    pyplot.close(fig)

    ##################################
    # MEAN-MEDIAN DIFFERENCE HISTOGRAM
    ##################################
    fig = pyplot.figure()
    pyplot.title('Mean-Median Difference Histogram')
    pyplot.hist(x=numpy.abs(flex.double(mean_map_vals)-flex.double(medn_map_vals)), bins=30, normed=True)
    pyplot.xlabel('Difference Map Value')
    pyplot.ylabel('Density')
    pyplot.tight_layout()
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-mean_median_diff_hist.png'.format(map_res)))
    pyplot.close(fig)

    ##################################
    # STATISTICAL MAP HISTOGRAMS
    ##################################
    fig = pyplot.figure()
    pyplot.title('Point Variation Map Histograms')
    # STANDARD DEVIATION MAPS
    pyplot.subplot(2, 1, 1)
    pyplot.hist(x=stds_map_vals, bins=n_bins)
    pyplot.xlabel('"Raw" Variation of Map Values')
    pyplot.ylabel('Count')
    # ADJUSTED STANDARD DEVIATION MAPS
    pyplot.subplot(2, 1, 2)
    pyplot.hist(x=sadj_map_vals, bins=n_bins)
    pyplot.xlabel('"Adjusted" Variation of Map Values')
    pyplot.ylabel('Count')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save both
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-stds_sadj_map_vals.png'.format(map_res)))
    pyplot.close(fig)

    ##################################
    # STD Values v ADJ STD Values
    ##################################
    fig = pyplot.figure()
    pyplot.title('Raw v Adjusted Map Variation')
    pyplot.scatter(x=stds_map_vals, y=sadj_map_vals)
    # Plot straight line between the min and max values
    min_val = min(stds_map_vals+sadj_map_vals)
    max_val = max(stds_map_vals+sadj_map_vals)
    pyplot.plot([min_val, max_val], [min_val, max_val], 'b--')
    # Axis labels
    pyplot.xlabel('"Raw" Variation of Map Values')
    pyplot.ylabel('"Adjusted" Variation of Map Values')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-std_v_adj_std_scatter.png'.format(map_res)))
    pyplot.close(fig)

    ########################################################

    pandda.log('=> Writing Map Uncertainties')

    # MAP PARAMS
    fig = pyplot.figure()
    pyplot.title('Map Statistics Histrograms')
    # MAP UNCERTAINTIES
    pyplot.hist(x=map_uncties, bins=n_bins, range=(min(map_uncties)-0.1,max(map_uncties)+0.1))
    pyplot.xlabel('Dataset Map Uncertainties')
    pyplot.ylabel('Count')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-d_map_uncertainties.png'.format(map_res)))
    pyplot.close(fig)

    ########################################################

    pandda.log('=> Scatter Plots')

    # MAP RESOLUTION V UNCERTAINTY
    fig = pyplot.figure()
    pyplot.title('High Resolution Limit v Map Uncertainty')
    pyplot.scatter(x=high_res, y=map_uncties)
    pyplot.xlabel('Dataset Resolution')
    pyplot.ylabel('Dataset Map Uncertainty')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-resolution_v_uncertainty.png'.format(map_res)))
    pyplot.close(fig)

    # MAP RESOLUTION V UNCERTAINTY
    fig = pyplot.figure()
    pyplot.title('High Resolution Limit v R-Free')
    pyplot.scatter(x=high_res, y=rfree)
    pyplot.xlabel('Dataset Resolution')
    pyplot.ylabel('Dataset R-Free')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-resolution_v_rfree.png'.format(map_res)))
    pyplot.close(fig)

    # RFREE V UNCERTAINTY
    fig = pyplot.figure()
    pyplot.title('R-Free v Uncertainty')
    pyplot.scatter(x=rfree, y=map_uncties)
    pyplot.xlabel('Dataset R-Free')
    pyplot.ylabel('Dataset Map Uncertainty')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-rfree_v_uncertainty.png'.format(map_res)))
    pyplot.close(fig)

    ########################################################

    pandda.log('=> Z-Map Distribution')

    # R-FACTORS
    fig = pyplot.figure()
    pyplot.title('Z-Map Statistics Histograms')
    # RFree
    pyplot.subplot(2, 1, 1)
    pyplot.hist(x=z_map_mean, bins=n_bins, range=(min(z_map_mean)-0.1,max(z_map_mean)+0.1))
    pyplot.xlabel('Z-Map Mean')
    pyplot.ylabel('Count')
    # RWork
    pyplot.subplot(2, 1, 2)
    pyplot.hist(x=z_map_std, bins=n_bins, range=(0, max(z_map_std)+0.1))
    pyplot.xlabel('Z-Map Std')
    pyplot.ylabel('Count')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save both
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-z_map_statistics.png'.format(map_res)))
    pyplot.close(fig)

    # Z-MAP SKEW V UNCERTAINTY
    fig = pyplot.figure()
    pyplot.title('Z-Map Normality Plots')
    pyplot.scatter(x=z_map_skew, y=z_map_kurt)
    pyplot.xlabel('Skew')
    pyplot.ylabel('Kurtosis')
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
    pyplot.savefig(os.path.join(img_out_dir, '{!s}A-z_map_skew_v_kurtosis.png'.format(map_res)))
    pyplot.close(fig)

def write_occupancy_graph(x_values, global_values, local_values, diff_values, out_file):
    """Write output graph from occupancy estimation"""

    # Get the x-value of the maximum difference
    max_x = x_values[diff_values.index(max(diff_values))]

    fig, (axis_1_1, axis_2_1) = pyplot.subplots(2, sharex=True)
    # 1st Plot - 1st Y-Axis
    line_1_1, = axis_1_1.plot(x_values, global_values, 'g--', label='GLOBAL CORRELATION', linewidth=2)
    line_1_2, = axis_1_1.plot(x_values, local_values, 'k--', label='LOCAL CORRELATION', linewidth=2)
    axis_1_1.set_ylabel('CORRELATION\nTO GROUND STATE', color='k', size=16)
    axis_1_1.set_ylim((-1, 1))
    # 2nd Plot - 1st Y-Axis
    line_2_1, = axis_2_1.plot(x_values, diff_values, 'b-', label='DIFFERENCE', linewidth=2)
    axis_2_1.set_ylabel('CORRELATION\nDIFFERENCE', color='k', size=16)
    axis_2_1.set_xlabel('1-BDC', color='k', size=16)
    axis_2_1.set_ylim((min(diff_values)-0.2, max(diff_values)+0.2))
    # Plot line at the maximum
    line_2_2, = axis_2_1.plot([max_x,max_x],[-1,1], 'k-', linewidth=2)
    text_2_1 = axis_2_1.text(0.02+max_x, 0.0, 'BDC='+str(1-max_x), size=14)
    # Joint legend
    axis_1_1.legend(handles=[line_1_1, line_1_2, line_2_1], loc=4, fontsize=16)
#    # Title
#    axis_2_1.set_title('Estimating Occupancy by Correlation Gradient Differences')
    # Remove spacing between subplots
    pyplot.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
    pyplot.setp([axis_2_1.get_xticklabels()], visible=True)
    # Apply tight layout to prevent overlaps
    pyplot.tight_layout()
    # Save
#    pyplot.savefig(d_handler.output_handler.get_file('occ_corr_png').format(event_key))
    pyplot.savefig(out_file)
    pyplot.close(fig)



# Contains all non-specialist code for the analysis of protein and ligand structures
# Does NOT require cctbx

import os, sys

from bamboo.common.errors import *

def div_bar(width=30):
    if width < 5: width=5
    return '#'*2 + '='*(width-5) + '>'*3




import os, time

from bamboo.wrappers.ligand_fitters import FitterObject

class RhofitObject(FitterObject):

    def _custom_init(self):
        self.program = 'rhofit'
        self.name = 'rhofit'

    def _custom_prepare_to_fit_ligand(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):
        """Prepare custom settings before fitting the ligand"""

        # Run the custom parts of this function
        self.workdir = os.path.join(outdir,'rhofit')

    def _create_program_arguments(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):
        """Create command line and standard input to fit the ligand"""

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = ['-l', ligcif, '-lp', ligpdb, '-m', mtz, '-p', apopdb, '-d', self.workdir]
        cmd_line_args.extend(cmd_line_flags)
        # Form Standard Input
        std_inpt_args = []
        std_inpt_args.extend(std_inpt_flags)

        return cmd_line_args, std_inpt_args

    def _post_process(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):

        # Get the output models
        self._find_output_files()

        return

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = []
        std_inpt_flags = []

        if '-keepHydrogens' in flags:
            cmd_line_flags.append('-keepH')

        return cmd_line_flags, std_inpt_flags

    def _find_output_files(self):
        """Reads the Rhofit summary to find the output files"""

        workdir = self.workdir
        summary = os.path.join(self.workdir, 'results.txt')

        if not os.path.exists(summary):
            raise FittingError('No Fitting Summary File Found: {!s}\n\t\tLog: {!s}'.format(summary, self.outlog))

        summary_contents = open(summary,'r').readlines()

        fields = []
        rows = []
        data = {}

        for i, line in enumerate(summary_contents):
            # Remove newlines and whitespace
            line = line.strip()
            if line.startswith('========================'):
                assert summary_contents[i-3].strip() == 'rhofit           ligand LigProt  Poorly', 'RHOFIT SCORE FILE NOT IN EXPECTED FORMAT! {!s}'.format(summary)
                assert summary_contents[i-2].strip() == 'total   Correl  strain contact fitting   LigProt contact to residues', 'RHOFIT SCORE FILE NOT IN EXPECTED FORMAT! {!s}'.format(summary)
                assert summary_contents[i-1].strip()   == 'File               Chain    score   coeff    score   score   atoms   (% means zero weighted in score)', 'RHOFIT SCORE FILE NOT IN EXPECTED FORMAT! {!s}'.format(summary)
                fields = ['File','Chain','Rhofit Score','RSCC','Strain Score','Contact Score','Poor Atoms','Contacts']
            elif fields and line:
                values = line.split()
                if len(values)>6:
                    # Add non-blank lines to the lines to process
                    rows.append(line.split())

        if not (fields and rows):
            raise FittingError('No Fitted Files Found in Summary File: {!s}\n\t\tLog: {!s}'.format(summary, self.outlog))

        # Order by rhofit score
        for i, row in enumerate(sorted(rows, key=lambda entr: float(entr[2]))):
            # Sort out the last column (splits into many columns)
            proc_row = row[0:7]
            proc_row.append(';'.join(row[7:]))
            # Create dictionary of score information
            rowdict = dict(zip(fields,proc_row))
            # Create a rank (the files are already ordered by a rhofit score)
            rank = i+1
            # Get the cluster, dictionary number, and rank within the cluster from the filename
            cluster, dictionary_num, cluster_rank = proc_row[0].replace('.pdb','').replace('Hit_','').split('_')
            # Add a couple more fields to the dict
            rowdict['Rank'] = rank
            rowdict['Cluster'] = int(cluster)
            rowdict['Cluster Rank'] = int(cluster_rank)
            rowdict['Dictionary Number'] = int(dictionary_num)
            # Copy the File to Path for Neatness
            rowdict['Path'] = os.path.join(workdir,rowdict['File'])
            # Create a common output score
            rowdict['InternalScore'] = rowdict['Rhofit Score']
            # Check path exists
            assert os.path.exists(rowdict['Path']), 'Rhofit output File does not exist: {!s}'.format(rowdict)

            # Add to list of output model data (ordered by rank from rhofit)
            data[rowdict['Path']] = rowdict
            # Add path to list of output models (ordered by rank from rhofit)
            self.outmodels.append(rowdict['Path'])

        # Store scoring data
        self.fittingdata = data

        return



import os

from bamboo.wrappers.ligand_builders import BuilderObject

class ProdrgObject(BuilderObject):

    def _custom_init(self):
        self.program = 'cprodrg'
        self.name = 'prodrg'

    def _create_program_arguments(self, ligsmile, outdir, outfile, flags):
        """Prepare the arguments for Prodrg to build the ligand"""

        # Create input smiles file
        self.infile = self.outtemplate+'.smile'
        self.outmol = self.outtemplate+'.mol'
        # Write smile file
        with open(self.infile, 'w') as file:
            file.write(ligsmile+'\n')
            file.write('CPNAME '+self.ligname)

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = ['XYZIN',self.infile,'LIBOUT',self.outcif,'XYZOUT',self.outpdb,'MOLOUT',self.outmol]
        cmd_line_args.extend(cmd_line_flags)
        # Form Std Input
        std_inpt_args = ['MINI YES']
        std_inpt_args.extend(std_inpt_flags)
        std_inpt_args.append('END')

        return cmd_line_args, std_inpt_args

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = []
        std_inpt_flags = []

        if 'noHydrogens' in flags:
            raise SystemExit('AAAHHH - FLAG NOT PROGRAMMED FOR PRODRG - `noHydrogens`')

        return cmd_line_flags, std_inpt_flags



import os, sys, glob

import libtbx.phil

from bamboo.common.command import CommandManager
from bamboo.common.path import rel_symlink

#######################################

blank_arg_prepend = {   '.mtz':'mtz=',
                        '.pdb':'pdb=',
                        '.cif':'cif=',
                        '.params':'params=' }

master_phil = libtbx.phil.parse("""
input {
    pdb = None
        .type = str
    mtz = 'refine.mtz'
        .type = str
    cif = None
        .type = str
        .multiple = True
    params = None
        .help = "File that contains occupancy grouping definitions and commands. Generated by the giant.create_occupancy_params script."
        .type = str
}
settings {
    program = *phenix refmac
        .type = choice
    create_occupancy_params = False
        .help = "Regenerate input.params file for occupancy refinement before and after refinement? Must also pass input.params file."
        .type = bool
    args = None
        .help = "Pass any additional arguments to the program? (Command Line arguments). Pass as quoted strings."
        .type = str
        .multiple = True
}
phenix {
    ordered_solvent = False
        .type = bool
    anisotropic = False
        .type = bool
    no_coords = False
        .type = bool
}
output {
    dir_prefix = 'Refine_'
        .type = str
    out_prefix = 'refine'
        .type = str
    link_prefix = 'refine'
        .type = str
}

""")

#######################################

def run(params):

    assert params.input.pdb is not None, 'No PDB given for refinement'
    assert params.input.mtz is not None, 'No MTZ given for refinement'
    if params.settings.create_occupancy_params: assert params.input.params, 'settings.create_occupancy_params is True, but you must also define input.params.'

    ########################
    current_dirs = sorted(glob.glob(params.output.dir_prefix+'*'))
    if not current_dirs:
        out_dir = params.output.dir_prefix+'1'
    else:
        current_dirs = [s.replace(params.output.dir_prefix, '') for s in current_dirs]
        out_dir = params.output.dir_prefix + str(sorted(map(int, current_dirs))[-1]+1)
    print 'Outputting to {}'.format(out_dir)
    os.mkdir(out_dir)

    ########################
    # Link input
    rel_symlink(params.input.pdb, os.path.abspath(os.path.join(out_dir, 'input.pdb')))
    rel_symlink(params.input.mtz, os.path.abspath(os.path.join(out_dir, 'input.mtz')))

    ########################
    output_prefix = os.path.join(out_dir, params.output.out_prefix)
    print 'Output Real Prefix = {}'.format(output_prefix)
    print 'Output Link Prefix = {}'.format(params.output.link_prefix)

    ########################
    # Create occupancy parameters
    if params.settings.create_occupancy_params and params.input.params:
        print 'Running giant.create_occupancy_params'
        print 'Generating occupancy model for the input pdb: {}'.format(params.input.pdb)
        if params.settings.program == 'phenix':
            os.system('giant.create_occupancy_params pdb={} phenix_occ_out={} refmac_occ_out=None verbose=False'.format(params.input.pdb, params.input.params))
        elif params.settings.program == 'refmac':
            os.system('giant.create_occupancy_params pdb={} phenix_occ_out=None refmac_occ_out={} verbose=False'.format(params.input.pdb, params.input.params))

    ########################
    # Phenix
    if params.settings.program == 'phenix':
        ########################
        # Build command string
        cm = CommandManager('phenix.refine')
        cm.add_command_line_arguments([ params.input.pdb, params.input.mtz ])
        cm.add_command_line_arguments([ 'output.prefix={}'.format(output_prefix) ])
        if params.input.cif:
            cm.add_command_line_arguments( params.input.cif )
        if params.input.params and os.path.exists(params.input.params):
            cm.add_command_line_arguments([ params.input.params ])
        if params.phenix.ordered_solvent:
            cm.add_command_line_arguments([ 'ordered_solvent=True' ])
        if params.phenix.anisotropic:
            cm.add_command_line_arguments([ 'adp.individual.anisotropic="not element H"' ])
        if params.phenix.no_coords:
            cm.add_command_line_arguments([ 'strategy=occupancies+individual_adp' ])

    elif params.settings.program == 'refmac':
        ########################
        # Build command string
        cm = CommandManager('refmac5')
        cm.add_command_line_arguments( ['xyzin', params.input.pdb, 'hklin', params.input.mtz] )
        cm.add_command_line_arguments( ['xyzout', output_prefix+'.pdb', 'hklout', output_prefix+'.mtz'] )
        if params.input.cif:
            for cif in params.input.cif:
              cm.add_command_line_arguments( ['libin', cif] )

        # Read in occupancy params
        if params.input.params:
            cm.add_standard_input( open(params.input.params).read().split('\n') )

        cm.add_standard_input( ['END'] )

    # Pass additional command line arguments?
    if params.settings.args:
        cm.add_command_line_arguments( params.input.args )

    ########################
    # Print + Run
    cm.print_settings()
    out = cm.run()
    print cm.output

    if out != 0:
        print cm.error

    ########################
    # Write log
    with open(output_prefix+'.quick.log', 'w') as fh:
        fh.write('\n\nSTDOUT\n\n')
        fh.write(cm.output)
        fh.write('\n\nSTDERR\n\n')
        fh.write(cm.error)

    ########################
    # Link output
    real_pdb = glob.glob(output_prefix+'*.pdb')
    link_pdb = params.output.link_prefix+'.pdb'
    if real_pdb:
        real_pdb = real_pdb[0]
        if os.path.exists(link_pdb) and os.path.islink(link_pdb):
            os.unlink(link_pdb)
        if (not os.path.exists(link_pdb)):
            os.symlink(real_pdb, link_pdb)

    real_mtz = glob.glob(output_prefix+'*.mtz')
    link_mtz = params.output.link_prefix+'.mtz'
    if real_mtz:
        real_mtz = real_mtz[0]
        if os.path.exists(link_mtz) and os.path.islink(link_mtz):
            os.unlink(link_mtz)
        if (not os.path.exists(link_mtz)):
            os.symlink(real_mtz, link_mtz)

    ########################
    # Create occupancy parameters
    if os.path.exists(link_pdb) and params.settings.create_occupancy_params and params.input.params:
        print 'Running giant.create_occupancy_params'
        print 'Generating occupancy model for the input pdb: {}'.format(link_pdb)
        if params.settings.program == 'phenix':
            os.system('giant.create_occupancy_params pdb={} phenix_occ_out={} refmac_occ_out=None verbose=False'.format(link_pdb, params.input.params))
        elif params.settings.program == 'refmac':
            os.system('giant.create_occupancy_params pdb={} phenix_occ_out=None refmac_occ_out={} verbose=False'.format(link_pdb, params.input.params))

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, shutil, tempfile

from bamboo.common.command import CommandManager

def get_pdb_bfactor_summary(pdb_file):
    """Analyse the b-factors in a pdb file and return a dictionary with all of the averages, rms deviations and Z-scores for the b-factors in a structure"""

    program, contents = extract_bfactor_statistics(pdb_file)
    # Process the output table from baverage
    processed_contents = process_baverage_output(contents)
    # Process the stdout from the program
    processed_stdout = process_baverage_stdout(program.out)

    return program, processed_contents

def process_baverage_output(contents):
    """Process baverage output"""

    chain = ''
    resid = ''

    processed = {}

    for line in contents:
        if not line:
            continue
        elif line.strip().startswith('#'):
            chain = line.split()[2]
            assert len(chain)==1, 'Chain ID is more than one letter!'
        else:
            if not chain:
                raise Exception('No Chain ID has been assigned!')

            resid, m_av, m_rms, s_av, s_rms, a_av, a_rms = map(float, line.split())
            resid = int(resid)

            processed[(chain, resid)] = bFactorSummary(chain, resid, m_av, m_rms, s_av, s_rms, a_av, a_rms)

    return processed

def extract_bfactor_statistics(pdb_file):
    """Analyse the b-factors in a pdb file and return a dictionary with all of the averages, rms deviations and Z-scores for the b-factors in a structure"""

    # Create a temporary file for the output summary table and pdb_file
    temp_handle, temp_path = tempfile.mkstemp(suffix='.table', prefix='baverage_')
    temp_handle2, temp_path2 = tempfile.mkstemp(suffix='.pdb', prefix='baverage_')

    BAVERAGE = CommandManager('baverage')
    BAVERAGE.add_command_line_arguments('XYZIN',pdb_file,'RMSTAB',temp_path,'XYZOUT',temp_path2)
    BAVERAGE.add_standard_input(['END'])
    BAVERAGE.run()

    if not os.path.exists(temp_path):
        raise ExternalProgramError('BAVERAGE has failed to calculate b-factor summary statistics for {!s}'.format(pdb_file))

    # Process Table and Standard Out
    table_contents = open(temp_path, 'r').read().split('\n')

    if not table_contents:
        raise ExternalProgramError('BAVERAGE has failed to calculate b-factor summary statistics for {!s}'.format(pdb_file))
    else:
        os.remove(temp_path)
        os.remove(temp_path2)

    return BAVERAGE, table_contents




STRUCTURE_MASK_NAMES = [    'bad structure - non-identical structures',
                            'bad structure - different space group'    ]
CRYSTAL_MASK_NAMES   = [    'bad crystal - isomorphous crystal',
                            'bad crystal - isomorphous structure',
                            'bad crystal - space group',
                            'bad crystal - data quality',
                            'bad crystal - data correlation',
                            'bad crystal - rfree'  ]
REJECT_MASK_NAMES    = [    'rejected - total',
                            'rejected - crystal',
                            'rejected - structure',
                            'rejected - unknown'  ]
FLAG_MASK_NAMES      = [    'noisy zmap',
                            'analysed'  ]

DATASET_INFO_FIELDS  = [    'high_resolution',
                            'low_resolution',
                            'r_work',
                            'r_free',
                            'rmsd_to_reference',
                            'space_group',
                            'uc_a',
                            'uc_b',
                            'uc_c',
                            'uc_alpha',
                            'uc_beta',
                            'uc_gamma',
                            'uc_vol',
                            'rejection_reason'                ]
DATASET_MAP_FIELDS   = [    'analysed_resolution',
                            'map_uncertainty',
                            'obs_map_mean',
                            'obs_map_rms',
                            'z_map_mean',
                            'z_map_std',
                            'z_map_skew',
                            'z_map_kurt'            ]
DATASET_EVENT_FIELDS = [    'site_idx',
                            '1-BDC',
                            'z_peak',
                            'z_mean',
                            'cluster_size',
                            'x','y','z',
                            'refx','refy','refz'    ]
SITE_TABLE_FIELDS    = [    'centroid',
                            'num_events',
                            'nearest_residue 1',
                            'nearest_residue 2',
                            'nearest_residue 3',
                            'native_centroid',
                            'near_crystal_contacts'     ]

class PanddaMaskNames:
    structure_mask_names = STRUCTURE_MASK_NAMES
    crystal_mask_names   = CRYSTAL_MASK_NAMES
    reject_mask_names    = REJECT_MASK_NAMES
    flag_mask_names      = FLAG_MASK_NAMES
    custom_mask_names    = ['no_analyse', 'no_build']
    all_mask_names       = structure_mask_names + crystal_mask_names + reject_mask_names + flag_mask_names + custom_mask_names

class PanddaTableFields:
    all_dataset_fields     = DATASET_INFO_FIELDS
    all_dataset_map_fields = DATASET_MAP_FIELDS
    all_event_fields       = DATASET_EVENT_FIELDS
    all_site_fields        = SITE_TABLE_FIELDS

########################################################################

class PanddaDatasetFilenames:
    # Input data...
    input_structure        = '{!s}-pandda-input.pdb'
    input_data             = '{!s}-pandda-input.mtz'
    # Dataset Information
    dataset_info           = '{!s}-info.csv'
    dataset_log            = '{!s}.log'
    dataset_pickle         = 'dataset.pickle'
    # Structure Files...
    aligned_structure      = '{!s}-aligned-structure.pdb'
    symmetry_copies        = '{!s}-aligned-sym-contacts.pdb'
    # Modelled Structures...
    modelled_structure     = '{!s}-pandda-model.pdb'
    ensemble_structure     = '{!s}-ensemble-model.pdb'
    # Dataset Maps
    sampled_map            = '{!s}-aligned-map.ccp4'
    # Z-Maps...
    mean_diff_map          = '{!s}-difference-from-mean.ccp4'
    z_map                  = '{!s}-z_map.ccp4'
    z_map_naive            = '{!s}-z_map_naive.ccp4'
    z_map_naive_norm       = '{!s}-z_map_naive_normalised.ccp4'
    z_map_uncertainty      = '{!s}-z_map_uncertainty.ccp4'
    z_map_uncertainty_norm = '{!s}-z_map_uncertainty_normalised.ccp4'
    z_map_corrected        = '{!s}-z_map_adjusted.ccp4'
    z_map_corrected_norm   = '{!s}-z_map_adjusted_normalised.ccp4'
    # Events...
    event_map              = '{!s}-event_{!s}_1-BDC_{!s}_map.ccp4'
    # Masks...
    high_z_mask            = '{!s}-high_z_mask.ccp4'
    grid_mask              = '{!s}-masked_grid.ccp4'
    # Ligands...
    ligand_coordinates     = '{!s}-ligand.pdb'
    ligand_restraints      = '{!s}-ligand.cif'
    ligand_image           = '{!s}-ligand.png'
    # Native (Output) Maps
    native_obs_map         = '{!s}-observed.native.ccp4'
    native_z_map           = '{!s}-z_map.native.ccp4'
    native_event_map       = '{!s}-event_{!s}_1-BDC_{!s}_map.native.ccp4'
    # Misc...
    z_peaks_csv            = '{!s}-z_map_peaks.csv'
    pymol_script           = 'load_maps.pml'
    ccp4mg_script          = 'ccp4mg_{!s}_{!s}.py'
    ccp4mg_png             = 'blob_{!s}_{!s}.png'

class PanddaDatasetPNGFilenames:
    s_map_png                   = '{!s}-obsv_map_dist.png'
    d_mean_map_png              = '{!s}-diff_mean_map_dist.png'
    z_map_naive_png             = '{!s}-z_map_dist_naive.png'
    z_map_naive_norm_png        = '{!s}-z_map_dist_naive_normalised.png'
    z_map_uncertainty_png       = '{!s}-z_map_dist_uncertainty.png'
    z_map_uncertainty_norm_png  = '{!s}-z_map_dist_uncertainty_normalised.png'
    z_map_corrected_png         = '{!s}-z_map_dist_adjusted.png'
    z_map_corrected_norm_png    = '{!s}-z_map_dist_adjusted_normalised.png'
    z_map_qq_plot_png           = '{!s}-z_map_dist_qq_plot.png'
    bdc_est_png                 = '{!s}-event_{!s}_BDC_estimation.png'
    unc_qqplot_png              = '{!s}-uncertainty-qqplot.png'
    obs_qqplot_sorted_png       = '{!s}-mean-v-obs-sorted-qqplot.png'
    obs_qqplot_unsorted_png     = '{!s}-mean-v-obs-unsorted-plot.png'

class PanddaAnalyserFoldernames:
    pass

class PanddaAnalyserFilenames:
    initial_html            = 'pandda_initial.html'
    analyse_html            = 'pandda_analyse.html'
    analyse_site_graph      = 'pandda_analyse_sites_graph.png'
    analyse_site_graph_mult = 'pandda_analyse_sites_graph_{!s}.png'
    # Output csvs
    event_info              = 'pandda_analyse_events.csv'
    site_info               = 'pandda_analyse_sites.csv'
    dataset_info            = 'all_datasets_info.csv'
    dataset_map_info        = 'all_datasets_info_maps.csv'
    dataset_combined_info   = 'all_datasets_info_combined.csv'
    # Statistical Maps
    mean_map                = '{!s}A-mean_map.ccp4'
    medn_map                = '{!s}A-medn_map.ccp4'
    stds_map                = '{!s}A-stds_map.ccp4'
    sadj_map                = '{!s}A-sadj_map.ccp4'
    skew_map                = '{!s}A-skew_map.ccp4'
    kurt_map                = '{!s}A-kurt_map.ccp4'
    bimo_map                = '{!s}A-bimo_map.ccp4'
    # Reference Files
    reference_structure     = 'reference.pdb'
    reference_dataset       = 'reference.mtz'
    reference_on_origin     = 'reference.shifted.pdb'
    reference_symmetry      = 'reference.symmetry.pdb'
    # Pymol images
    pymol_sites_png_1       = 'pandda_analyse_sites_pymol_1.png'
    pymol_sites_png_2       = 'pandda_analyse_sites_pymol_2.png'
    pymol_sites_py          = 'pandda_analyse_sites_pymol.py'
    pymol_sites_pml         = 'pandda_analyse_sites_pymol.pml'

class PanddaInspectorFilenames:
    inspect_html            = 'pandda_inspect.html'
    inspect_site_graph      = 'pandda_inspect_sites_graph.png'
    inspect_site_graph_mult = 'pandda_inspect_sites_graph_{!s}.png'
    # output csvs
    event_info              = 'pandda_inspect_events.csv'
    site_info               = 'pandda_inspect_sites.csv'




import numpy
from math import cos, pi, sqrt
from scitbx.array_family import flex

def unit_cell_diagonals(a, b, c, alpha=90, beta=90, gamma=90):
    """Calculate the lengths of the diagonals in the unit cell (across the three faces from the origin)"""
    d_ab = sqrt(a**2 + b**2 - 2*a*b*cos((1-gamma/180.0)*pi) )
    d_bc = sqrt(b**2 + c**2 - 2*b*c*cos((1-alpha/180.0)*pi) )
    d_ca = sqrt(c**2 + a**2 - 2*c*a*cos((1-beta/180.0)*pi)  )
    return d_ab, d_bc, d_ca

def fractional_change_for_diagonals(uc1_diags, uc2_diags):
    """Calculate the fractional change between two sets of diagonals"""
    d_ab_1, d_bc_1, d_ca_1 = uc1_diags
    d_ab_2, d_bc_2, d_ca_2 = uc2_diags
    m_ab = abs(d_ab_1-d_ab_2)/min(d_ab_1, d_ab_2)
    m_bc = abs(d_bc_1-d_bc_2)/min(d_bc_1, d_bc_2)
    m_ca = abs(d_ca_1-d_ca_2)/min(d_ca_1, d_ca_2)
    return m_ab, m_bc, m_ca

#############################################################################

def lcv_from_diagonals(uc1_diags, uc2_diags):
    chg_diags = fractional_change_for_diagonals(uc1_diags=uc1_diags, uc2_diags=uc2_diags)
    return max(chg_diags)

def lcv_from_unit_cell_parameters(uc1_params, uc2_params):
    """Calculate the Linear Cell Variation for two sets of unit cell parameters (a,b,c,alpha,beta,gamma)"""
    uc1_diags = unit_cell_diagonals(*uc1_params)
    uc2_diags = unit_cell_diagonals(*uc2.params)
    return lcv_from_diagonals(uc1_diags=uc1_diags, uc2_diags=uc2_diags)

def lcv_from_unit_cells(uc1, uc2):
    """Calculate the Linear Cell Variation for two unit cells"""
    return lcv_from_unit_cell_parameters(uc1_params=uc1.parameters(), uc2_params=uc2.parameters())

#############################################################################

def pairwise_lcv(unit_cells=None, parameters=None):
    """Calculate the LCV for all pairs of unit cells/parameters"""
    assert [unit_cells, parameters].count(None) == 1, "Provide either unit_cells or parameters"
    if unit_cells:  parameters = [uc.parameters() for uc in unit_cells]
    num_objs = len(parameters)
    diags = numpy.empty([num_objs], dtype=object)
    dist  = numpy.empty([num_objs]*2, dtype=object)
    for i_uc1, uc1 in enumerate(parameters):
        diags[i_uc1] = unit_cell_diagonals(*uc1)
    for i_uc1, i_uc2 in flex.nested_loop((num_objs, num_objs)):
        dist[i_uc1, i_uc2] = lcv_from_diagonals(uc1_diags=diags[i_uc1], uc2_diags=diags[i_uc2])
    return dist

def pairwise_ecv(unit_cells=None, parameters=None):
    """Calculate the Euclidean Cell Variation (ECV) for all pairs of unit cells/parameters"""
    assert [unit_cells, parameters].count(None) == 1, "Provide either unit_cells or parameters"
    if unit_cells:  parameters = [uc.parameters() for uc in unit_cells]
    num_objs = len(parameters)
    diags = numpy.empty([num_objs], dtype=object)
    dist  = numpy.empty([num_objs]*2, dtype=object)
    for i_uc1, uc1 in enumerate(parameters):
        diags[i_uc1] = unit_cell_diagonals(*uc1)
    for i_uc1, i_uc2 in flex.nested_loop((num_objs, num_objs)):
        dist[i_uc1, i_uc2] = flex.double(fractional_change_for_diagonals(diags[i_uc1],diags[i_uc2])).norm()
    return dist



import os,sys,re

from bamboo.common.errors import *
from bamboo.rdkit_utils.smile import check_smile_readable

from rdkit.Chem import AllChem as Chem

return_codes = {0 : 'Reference {!s} matches the query {!s} as far as I can tell!',
                1 : 'Reference {!s} has no matching substructure with query {!s}',
                2 : 'Reference {!s} has more atoms than query {!s}',
                3 : 'Reference {!s} has fewer atoms than query {!s}',
                4 : 'Reference {!s} has more bonds than query {!s}',
                5 : 'Reference {!s} has fewer bonds than query {!s}',
                6 : 'Reference {!s} has a different charge to query {!s}',
                7 : 'Reference {!s} does not have a full substructure match to query {!s}'}

def check_pdb_readable(file):
    """Checks a file is readable by Rdkit. Returns a mol object if successful"""

    mol = Chem.MolFromPDBFile(file)
    if not mol: raise LigandCheckError('File not rdkit-readable: {!s}'.format(file))
    else:       return mol

def has_bond_orders(mol):
    """Checks to see if the given mol has bond orders assigned"""

    if len([1 for b in mol.GetBonds() if b.GetBondTypeAsDouble() != 1.0]):
        return True
    else:
        return False

def check_and_assign_bond_orders_dynamic(mol1, mol2):
    """Checks both mol1 and mol2 to see if they have bond orders. If one does, but the other does not, then it attempts to transfer bond orders appropriately"""

    try:
        if has_bond_orders(mol1) and has_bond_orders(mol2):
            # Both have bond orders - do nothing
            pass
        elif has_bond_orders(mol1) and (not has_bond_orders(mol2)):
            # mol1 has, but mol2 does not - transfer from mol1 -> mol2
            mol2 = Chem.AssignBondOrdersFromTemplate(mol1, mol2)
        elif (not has_bond_orders(mol1)) and has_bond_orders(mol2):
            # mol2 has, but mol1 does not - transfer from mol2 -> mol1
            mol1 = Chem.AssignBondOrdersFromTemplate(mol2, mol1)
        else:
            # Neither has bond orders - pass
            pass
    except ValueError as err:
        raise BondAssignmentError('Failed to assign bond orders: {!s}'.format(err))

    if not mol1:
        raise Exception('Mol1 does not exist!')
    elif not mol2:
        raise Exception('Mol2 does not exist!')

    return mol1, mol2

def check_and_assign_bond_orders_static(mol1, mol2):
    """Transfers bond orders from mol1 to mol2, if bond orders are present"""

    try:
        # transfer from mol1 -> mol2
        mol2 = Chem.AssignBondOrdersFromTemplate(mol1, mol2)
    except ValueError as err:
        raise BondAssignmentError('Failed to assign bond orders: {!s}'.format(err))

    if not mol1:
        raise Exception('Mol1 does not exist!')
    elif not mol2:
        raise Exception('Mol2 does not exist!')

    return mol1, mol2

def check_ligands_from_file_identical(ref, query, dynamicBondAssignment=True):
    """Check two ligands (pdb files) are identical - gives similarity relative to the first file"""
    rc = check_ligand_mols_identical(check_pdb_readable(ref), check_pdb_readable(query), dynamicBondAssignment=dynamicBondAssignment)
    return rc, return_codes[rc].format('file','file')

def check_ligand_from_file_identical_to_smile(smile, query, dynamicBondAssignment=False):
    """Checks ligand (pdb file) is the same as the smile"""
    rc = check_ligand_mols_identical(check_smile_readable(smile), check_pdb_readable(query), dynamicBondAssignment=dynamicBondAssignment)
    return rc, return_codes[rc].format('smile','file')

def check_ligand_identical_to_smile(smile, mol, dynamicBondAssignment=False):
    """Checks ligand (mol) is the same as the smile"""
    rc = check_ligand_mols_identical(ref=check_smile_readable(smile), mol=mol, dynamicBondAssignment=dynamicBondAssignment)
    return rc, return_codes[rc].format('smile','mol')

def check_ligand_mols_identical(ref, mol, dynamicBondAssignment=True):
    """Check two ligands (mols) are identical - gives similarity relative to the first mol"""

    # Assign bond orders (and return if fails)
    try:
        if dynamicBondAssignment:
            # Keeps existing bond orders, or transfers if not present
            ref, mol = check_and_assign_bond_orders_dynamic(ref, mol)
        else:
            # Assigns bond orders from ref -> mol
            ref, mol = check_and_assign_bond_orders_static(ref, mol)
    except BondAssignmentError:
        return 1

    # Check Number of Atoms
    if ref.GetNumAtoms() < mol.GetNumAtoms():
        return 2
    if ref.GetNumAtoms() > mol.GetNumAtoms():
        return 3
    # Check Number of Bonds
    if ref.GetNumBonds() < mol.GetNumBonds():
        return 4
    if ref.GetNumBonds() > mol.GetNumBonds():
        return 5
    # Check Charge (This is where a lot of programs go 'wrong')
    if Chem.GetFormalCharge(ref) != Chem.GetFormalCharge(mol):
        return 6
#    try:
    # Check Substructure Matches
    if ref.GetNumAtoms() != len(mol.GetSubstructMatch(ref)):
        return 7
#    except ValueError as err:
#        raise LigandCheckError(' Value Error in GetSubstructMatch: {!s}'.format(err))

    return 0




class mask_collection(object):
    def __init__(self):
        """Stores lots of similar mask objects. Masks can be combined in multiple ways"""
        self.masks = {}
        self._mask_length = None
        self._entry_ids = None

    def _invert_mask(self, mask):
        """Invert a list of boolean values"""
        return [not m for m in mask]

    def set_mask_length(self, mask_length):
        """Set the mask length"""
        self._mask_length = mask_length

    def set_entry_ids(self, entry_ids):
        assert len(entry_ids) == self._mask_length
        self._entry_ids = entry_ids

    def add_mask(self, mask_name, mask):
        """Add a new mask"""
        assert len(mask) == self._mask_length
        self.masks[mask_name] = [True if m else False for m in mask]

    def get_mask(self, mask_name, invert=False):
        """Get a single mask"""
        mask = self.masks[mask_name]
        if invert: return self._invert_mask(mask)
        else:      return mask

    def has_mask(self, mask_name):
        """Check if it has a single mask"""
        return mask_name in self.masks.keys()

    def set_mask_value(self, mask_name, entry_id, value):
        """Set a particular entry in the mask `mask_name`, corresponding to `entry_id` to `value`"""
        assert mask_name in self.masks.keys()
        assert entry_id in self._entry_ids
        self.masks[mask_name][self._entry_ids.index(entry_id)] = bool(value)

    def get_mask_value(self, mask_name, entry_id):
        """Set a particular entry in the mask `mask_name`, corresponding to `entry_id` to `value`"""
        assert mask_name in self.masks.keys()
        assert entry_id in self._entry_ids
        return self.masks[mask_name][self._entry_ids.index(entry_id)]

    def mask(self, mask_name, input_list, invert=False):
        """Mask input_list according to the mask with name mask_name. Invert mask with `invert`"""
        assert isinstance(invert, bool)
        return [l for m, l in zip(self.masks[mask_name], input_list) if m != invert]

    def mask_custom(self, mask, input_list, invert=False):
        """Mask input_list according to the mask. Invert mask with `invert`"""
        assert isinstance(invert, bool)
        assert len(mask) == self._mask_length, 'Given mask is the wrong length'
        return [l for m, l in zip(mask, input_list) if m != invert]

    def combine_masks(self, mask_names, invert=False):
        """Combine lots of different masks. Invert output mask with `invert`"""
        masks = [self.get_mask(m_name) for m_name in mask_names]
        return self.combine_masks_custom(masks, invert=invert)

    def combine_masks_custom(self, masks, invert=False):
        """Combine a list of masks. Invert output mask with `invert`"""
        combined_mask = [max([m[i] for m in masks]) for i in xrange(self._mask_length)]
        if invert: return self._invert_mask(combined_mask)
        else: return combined_mask




import os

from scitbx.array_family import flex

from giant.html import GIANT_HTML_ENV
from bamboo.html import path2url

def write_html_summary(fname, atom_group_summaries):

    # Get template to be filled in
    template = GIANT_HTML_ENV.get_template('summary_page.html')

    # ===========================================================>
    # Construct the data object to populate the template
    output_data = {}
    output_data['header'] = 'Structure Summary'
    output_data['title'] = 'Structure Summary'

#    # ===========================================================>
#    # Progress Bars
#    output_data['progress_bar'] = []
#    output_data['progress_bar'].append({'title':'Dataset Summary', 'data':[]})
#    output_data['progress_bar'][0]['data'].append({'text':'Interesting',     'colour':'success', 'size':100.0*num_interesting/len_data})
#    output_data['progress_bar'][0]['data'].append({'text':'Not Interesting', 'colour':'info',    'size':100.0*num_not_interesting/len_data})
#    output_data['progress_bar'][0]['data'].append({'text':'Not Analysed',    'colour':'warning', 'size':100.0*num_not_analysed/len_data})
#    output_data['progress_bar'][0]['data'].append({'text':'Rejected',        'colour':'danger',  'size':100.0*num_rejected/len_data})

    # ===========================================================>
    # Tables
    output_data['table'] = {}
    output_data['table']['column_headings'] = [ 'Num Atoms',

                                                'Min BFactor',
                                                'Max BFactor',
                                                'Mean BFactor',
                                                'RMS BFactor',

                                                'Min BFactor-Z',
                                                'Max BFactor-Z',
                                                'Mean BFactor-Z',
                                                'RMS BFactor-Z',
                                            ]
    output_data['table']['rows'] = []

    for ag_sum in atom_group_summaries:
        # Extract the atom group info as a dictionary
        ed_sum = ag_sum.edstats_scores

        bfs      = ag_sum.b_factors
        bfs_stat = ag_sum.b_factors_statistics
        bfz      = ag_sum.b_factors_z
        bfz_stat = ag_sum.b_factors_z_statistics

        # colour choices - 'success', 'muted', 'danger'
        # icon choices   - 'ok', 'flag', 'remove'

        columns = []
        # Num Atoms
        columns.append({'colour':'default', 'message':round(bfs.size(),1)})
        # B-Factors
        columns.append({'colour':'default', 'message':round(bfs_stat.min,1)})
        columns.append({'colour':'default', 'message':round(bfs_stat.max,1)})
        columns.append({'colour':'default', 'message':round(bfs_stat.mean,1)})
        columns.append({'colour':'default', 'message':round((flex.sum_sq(bfs)/bfs.size())**0.5,1)})

        # B-Factors Z-Scores
        columns.append({'colour':'default', 'message':round(bfz_stat.min,2)})
        columns.append({'colour':'default', 'message':round(bfz_stat.max,2)})
        columns.append({'colour':'default', 'message':round(bfz_stat.mean,2)})
        if (flex.sum_sq(bfz)/bfz.size())**0.5 < 2.5:
            columns.append({'colour':'default', 'message':round((flex.sum_sq(bfz)/bfz.size())**0.5,1)})
        else:
            columns.append({'colour':'danger', 'icon':'remove', 'message':round((flex.sum_sq(bfz)/bfz.size())**0.5,1)})

        row_message = '' if (flex.sum_sq(bfz)/bfz.size())**0.5 < 2.5 else \
                      'Check'
        row_colour  = 'success' if row_message == '' else \
                      'danger'

        output_data['table']['rows'].append({'heading' : ag_sum.atom_group.id_str(),
                                             'message' : row_message,
                                             'colour'  : row_colour,
                                             'columns' : columns})

    with open(fname, 'w') as out_html:
        out_html.write(template.render(output_data))



import os

from libtbx import easy_pickle

from bamboo.common.logs import Log

class Program(object):
    """Class meant to provide basic functionality for programs and pipelines"""

    _NAME = None
    _TEXT = None
    _VERSION = None

    _allowed_statuses = ['running','done','errored']

    log = Log()

    def write_running_parameters_to_log(self):
        self.log('===================================>>>', True)
        self.log('PROCESSED ARGUMENTS', True)
        self.log('===================================>>>', True)
        self.log(self.master_phil.format(python_object=self._input_params).as_str(), True)

        self.log('===================================>>>', True)
        self.log('ARGUMENTS DIFFERENT TO DEFAULTS', True)
        self.log('===================================>>>', True)
        self.log(self.master_phil.fetch_diff(source=self.master_phil.format(python_object=self._input_params)).as_str(), True)

    def check_for_matplotlib(self):
        """Check to see whether we can load matplotlib"""
        try:
            import matplotlib
            matplotlib.use('Agg')
            # Setup so that we can write without a display connected
            matplotlib.interactive(0)
            # Find the backend
            default_backend, validate_function = matplotlib.defaultParams['backend']
            self.log('===================================>>>')
            self.log('MATPLOTLIB LOADED. Using Backend: {!s}'.format(default_backend))
            from matplotlib import pyplot
            # Test non-interactive-ness
            assert not pyplot.isinteractive(), 'PYPLOT IS STILL IN INTERACTIVE MODE'
            # Set the style to ggplot (the prettiest)
            pyplot.style.use('ggplot')
            self.log('PYPLOT loaded successfully')
            self.log('===================================>>>', True)
            self.log('', True)
            return True
        except:
            self.log('===================================>>>', True)
            self.log('>> COULD NOT IMPORT MATPLOTLIB. CANNOT GENERATE GRAPHS.', True)
            self.log('===================================>>>', True)
            self.log('', True)
            return False

    def update_status(self, status):
        """Set log files to indicate the status of the program"""

        assert status in self._allowed_statuses
        # Delete any that may exist
        existing_files = [self.output_handler.get_file('status').format(f) for f in self._allowed_statuses]
        [os.remove(f) for f in existing_files if os.path.exists(f)]
        # Create the new  status file
        with open(self.output_handler.get_file('status').format(status), 'w') as fh: fh.write('')

    def pickle(self, pickle_file, pickle_object, overwrite=True):
        """Takes an object and pickles it"""
        if os.path.exists(pickle_file) and not overwrite:
            self.log('NOT PICKLING: {!s}'.format(os.path.relpath(pickle_file, start=self.out_dir)))
        else:
            self.log('Pickling Object: {!s}'.format(os.path.relpath(pickle_file, start=self.out_dir)))
            easy_pickle.dump(pickle_file, pickle_object)

    def unpickle(self, pickle_file):
        """Takes an object and unpickles it"""
        self.log('Unpickling File: {!s}'.format(os.path.relpath(pickle_file, start=self.out_dir)))
        return easy_pickle.load(pickle_file)




import os, sys, copy, re, shutil

import libtbx.phil
import libtbx.easy_mp
from bamboo.common.command import CommandManager
from giant.xray.data import CrystalSummary

#######################################

blank_arg_prepend = {'.pdb':'reference_pdb=', '.mtz':'mtz='}

master_phil = libtbx.phil.parse("""
input {
    reference_pdb = None
        .type = path
        .multiple = True
    reference_mtz = None
        .type = path
    mtz = None
        .type = path
        .multiple = True
}
options {
    use_reference_rfree_flags = False
        .type = bool
    cpus = 1
        .type = int
}
output {
    overwrite = True
        .type = bool
    dir_suffix = '_dimple'
        .type = str
}
""")

#######################################

def run(params):

    def run_dimple(c):
        print '============================>'
        print 'Running New Dimple'
        c.print_settings()
        ret_code = c.run()
        if ret_code != 0:
            print '============================>'
            print 'Dimple returned with an error'
            print '============================>'
            c.print_settings()
            print '============================>'
            print c.output
            print '============================>'
            print c.error
            print '============================>'
        return ret_code

    ref_mtz_file = params.input.reference_mtz
    if ref_mtz_file: ref_mtz_file = os.path.abspath(ref_mtz_file)

    for ref_pdb_file in params.input.reference_pdb:

        commands = []

        ref_pdb_file = os.path.abspath(ref_pdb_file)

        if params.options.use_reference_rfree_flags:
            if not ref_mtz_file: ref_mtz_file = os.path.abspath(ref_pdb_file.replace('.pdb','.mtz'))
            assert os.path.exists(ref_mtz_file), 'Reference MTZ does not exist: {}'.format(ref_mtz_file)

        output_dirname = os.path.basename(ref_pdb_file).replace('.pdb', params.output.dir_suffix)
        print 'Placing all dimple runs for {} into directories called {}'.format(ref_pdb_file, output_dirname)

        for mtz_file in params.input.mtz:

            mtz_file = os.path.abspath(mtz_file)

            output_dir = os.path.join(os.path.dirname(mtz_file), output_dirname)

            if os.path.exists(output_dir):
                if params.output.overwrite:
                    shutil.rmtree(output_dir)
                else:
                    raise Exception('Output Directory already exists: {}'.format(output_dir))

            # Need to precreate the directory
            os.mkdir(output_dir)

            cm = CommandManager('dimple')
            if params.options.use_reference_rfree_flags:
                cm.add_command_line_arguments([ '--free-r-flags', ref_mtz_file ])
            cm.add_command_line_arguments([ mtz_file,
                                            ref_pdb_file,
                                            output_dir,
                                        ])
            cm.print_settings()
            commands.append(cm)

        # Run Dimples for this reference pdb file
        print '============================>'
        print 'Running Dimple for {} Datasets'.format(len(commands))
        returned = libtbx.easy_mp.pool_map(fixed_func=run_dimple, args=commands, processes=params.options.cpus)

        print '============================>'
        print 'Return Code Total: {}'.format(sum(returned))

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)




import pandas
import iotbx.pdb

from scitbx.array_family import flex
from giant.maths.geometry import is_within

import iotbx.pdb.hierarchy as iotbx_pdbh

#---------------------------------------------------->

def make_label(obj):
    """Return the relevant label for a supplied hierarchy/atom object"""
    if isinstance(obj, iotbx_pdbh.residue_group):
        return label_from_residue_group(obj)
    elif isinstance(obj, iotbx_pdbh.atom_group):
        return label_from_atom_group(obj)
    elif isinstance(obj, iotbx_pdbh.conformer):
        return label_from_conformer(obj)
    elif isinstance(obj, iotbx_pdbh.residue):
        return label_from_residue(obj)
    elif isinstance(obj, iotbx_pdbh.atom):
        return label_from_atom(obj)
    elif isinstance(obj, iotbx_pdbh.atom_with_labels):
        return label_from_atom_with_labels(obj)
    else:
        raise Exception('Invalid object type provided: {}'.format(type(obj)))

#---------------------------------------------------->

def label_from_residue_group(rg):
    """Return (chain_id, resid)"""
    ch = rg.parent()
    return (ch.id, rg.resid())

def label_from_atom_group(ag):
    """Return (chain_id, resid, altloc)"""
    rg = ag.parent()
    ch = rg.parent()
    return (ch.id, rg.resid(), ag.altloc)

#---------------------------------------------------->

def label_from_conformer(cnf):
    """Return (chain_id, resid, altloc). Must only have one residue."""
    ch = cnf.parent()
    res = cnf.only_residue()
    return (ch.id, res.resid(), cnf.altloc)

def label_from_residue(res):
    """Return (chain_id, resid, altloc)."""
    cnf = res.parent()
    ch = cnf.parent()
    return (ch.id, res.resid(), cnf.altloc)

#---------------------------------------------------->

def label_from_atom(at):
    """Return (chain_id, resid, altloc)"""
    ag = at.parent()
    rg = ag.parent()
    ch = rg.parent()
    return (ch.id, rg.resid(), ag.altloc)

def label_from_atom_with_labels(at):
    """Return (chain_id, resid, altloc)"""
    return (at.chain_id, at.resid(), at.altloc)

#---------------------------------------------------->

def calculate_residue_group_occupancy(residue_group):
    """
    Calculate the occupancy of a whole residue, allowing for partial occupancy of partial conformers.
    If conformers are present, the maximum occupancies of each conformer are summed
    """
    rg = residue_group
    if [c.altloc for c in rg.conformers()] == ['']:
        res_occ = max(rg.atoms().extract_occ())
    else:
        res_occ = sum([max(c.atoms().extract_occ()) for c in rg.conformers() if c.altloc])
    if res_occ > 1:
        # Currently errors silently by returning None
        return None
    return res_occ

def calculate_residue_group_rmsd(residue_group_1, residue_group_2):
    """Calculate rmsd between two hierarchically-identical residue groups (only coordinates, occupancies, and b-factors allowed to be different)"""
    rg_1 = residue_group_1; rg_2 = residue_group_2
    ats_1 = rg_1.atoms();   ats_2 = rg_2.atoms()
    # Check that the residues are equivalent
    assert ats_1.size()                       == ats_2.size()
    assert list(ats_1.extract_name()   )      == list(ats_2.extract_name()   )
    assert list(ats_1.extract_element())      == list(ats_2.extract_element())
    assert [a.parent().altloc for a in ats_1] == [a.parent().altloc for a in ats_2]
    rmsd = (ats_1.extract_xyz()-ats_2.extract_xyz()).rms_length()
    return rmsd

def calculate_residue_group_bfactor_ratio(residue_group, hierarchy, data_table=None, rg_label=None, column_suffix=''):
    """Calculate bfactor quality metrics of the residue to surrounding residues"""

    rg_sel = residue_group
    # Set defaults
    if rg_label is None:    rg_label = (rg_sel.unique_resnames()[0]+'-'+rg_sel.parent().id+'-'+rg_sel.resseq+rg_sel.icode).replace(' ','')
    if data_table is None:  data_table = pandas.DataFrame(index=[rg_label], column=[])
    # Check validity
    if len(rg_sel.unique_resnames()) != 1: raise Exception(rg_label+': More than one residue name associated with residue group -- cannot process')

    # Extract rg_sel objects
    rg_sel_ags    = rg_sel.atom_groups()
    rg_sel_atoms  = rg_sel.atoms()
    rg_sel_coords = rg_sel_atoms.extract_xyz()
    # Select nearby atom_group for scoring
    near_ags = [ag for ag in hierarchy.atom_groups() if (is_within(4, rg_sel_coords, ag.atoms().extract_xyz()) and (ag not in rg_sel_ags))]
    if near_ags:
        # Extract atoms from nearby groups
        near_ats = iotbx.pdb.hierarchy.af_shared_atom()
        [near_ats.extend(ag.detached_copy().atoms()) for ag in near_ags]
        # Calculate B-factors of the residue
        res_mean_b = flex.mean_weighted(rg_sel_atoms.extract_b(), rg_sel_atoms.extract_occ())
        data_table.set_value(   index = rg_label,
                                col   = 'Average B-factor (Residue)'+column_suffix,
                                value = res_mean_b )
        # Calculate B-factors of the surrounding atoms
        sch_mean_b = flex.mean_weighted(near_ats.extract_b(), near_ats.extract_occ())
        data_table.set_value(   index = rg_label,
                                col   = 'Average B-factor (Surroundings)'+column_suffix,
                                value = sch_mean_b )
        # Store the ratio of the b-factors
        data_table.set_value(   index = rg_label,
                                col   = 'Surroundings B-factor Ratio'+column_suffix,
                                value = res_mean_b/sch_mean_b )

    return data_table





import numpy

from scitbx.array_family import flex

def calculate_bdc_subtracted_map(ref_map, query_map, bdc):
    """Calculate query_map - bdc*ref_map"""
    assert isinstance(ref_map, flex.double), 'ref_map must of type flex.double'
    assert isinstance(query_map, flex.double), 'query_map must of type flex.double'
    assert ref_map.all() == query_map.all(), 'ref_map and query_map are not the same size: {!s} != {!s}'.format(ref_map.all(), query_map.all())
    assert ref_map.size() == query_map.size(), 'ref_map and query_map are not the same size: {!s} != {!s}'.format(ref_map.size(), query_map.size())
    # Important for subtraction to be "ref_map*bdc" so that the output is a flex.double array
    diff_map = query_map - ref_map*bdc
    return diff_map

def calculate_varying_bdc_correlations(ref_map, query_map, feature_region, min_remain=0.0, max_remain=1.0, bdc_increment=0.01, reference_region=None, verbose=False):
    """Estimate the background corrections of a feature in "query", defined by grid points "feature_region". Reference should be an equivalent map differing only by not having the feature"""

    # Check that the inputs are the right types
    assert isinstance(ref_map, flex.double), 'ref_map must of type flex.double'
    assert isinstance(query_map, flex.double), 'query_map must of type flex.double'
    assert isinstance(feature_region, list), 'grid_point_selection must be of type list'
    assert isinstance(reference_region, list), 'grid_point_selection must be of type list'
    # Check that the two maps are the same size
    assert ref_map.all() == query_map.all(), 'ref_map and query_map are not the same size: {!s} != {!s}'.format(ref_map.all(), query_map.all())
    assert ref_map.size() == query_map.size(), 'ref_map and query_map are not the same size: {!s} != {!s}'.format(ref_map.size(), query_map.size())

    ##########################################################################################

    # Calculate the map indices of the grid points where the feature is
    grid_accessor = ref_map.accessor()
    feature_region_idx = flex.size_t(map(grid_accessor, feature_region))
    # Calculate the map indices of the grid points of the reference region of the map
    if reference_region: reference_region_idx = flex.size_t(map(grid_accessor, reference_region))
    else:                reference_region_idx = None

    ##########################################################################################

    # Extract the reference map values around the site of interest
    feature_region_vals_ref = ref_map.as_1d().select(feature_region_idx)
    # Extract the reference map values for the correlation correction
    if reference_region: reference_region_vals_ref = ref_map.as_1d().select(reference_region_idx)
    else:                reference_region_vals_ref = ref_map.as_1d()

    ##########################################################################################

    return_vals = []

    # Create list of fractions to subtract
    all_feature_remains = numpy.arange(min_remain, max_remain, bdc_increment)
    # Ensure last value is 1.0
    if all_feature_remains[-1] != max_remain:
        all_feature_remains = numpy.arange(min_remain, max_remain+bdc_increment, bdc_increment)
        all_feature_remains[-1] = max_remain

    # Iterate through different amounts of reference map subtraction to estimate feature correction
    for i_remain, feature_remain in enumerate(all_feature_remains):

        # Calculate how much of the reference map to take away
        bdc = 1 - feature_remain

        # Calculate the background-subtracted difference map
        diff_map = calculate_bdc_subtracted_map(ref_map=ref_map, query_map=query_map, bdc=bdc)

        # Extract feature region values
        feature_region_vals_diff = diff_map.as_1d().select(feature_region_idx)
        # Extract reference region values
        if reference_region: reference_region_vals_diff = diff_map.as_1d().select(reference_region_idx)
        else:                reference_region_vals_diff = diff_map.as_1d()

        feature_region_corr     = flex.linear_correlation(feature_region_vals_diff, feature_region_vals_ref, subtract_mean=False)
        reference_region_corr   = flex.linear_correlation(reference_region_vals_diff, reference_region_vals_ref, subtract_mean=False)

        return_vals.append((feature_remain, feature_region_corr.coefficient(), reference_region_corr.coefficient()))

    return zip(*return_vals)

def calculate_maximum_series_discrepancy(labels, series_1, series_2):
    """Calculate the point at which two series are maximally different"""
    assert len(series_1) == len(series_2)
    assert len(series_1) == len(labels)
    diffs = [series_1[i] - series_2[i] for i in range(len(series_1))]
    return labels[diffs.index(max(diffs))]

def estimate_event_background_correction(ref_map, query_map, feature_region, min_remain=0.0, max_remain=1.0, bdc_increment=0.01, reference_region=None, method='gradient', verbose=False):
    """Calculate an estimate for the correction of a defined event"""
    # Check method is valid
    assert method in ['gradient', 'value']
    # Calculate the correlations to the reference map for a range of corrections
    event_remains, event_corrs, global_corrs = calculate_varying_bdc_correlations(
        ref_map          = ref_map,
        query_map        = query_map,
        feature_region   = feature_region,
        reference_region = reference_region,
        min_remain       = min_remain,
        max_remain       = max_remain,
        bdc_increment    = bdc_increment,
        verbose          = verbose)

    # Estimate the event corrections (1-bdc)
    if method == 'value':
        # Calculate the maximum discrepancy in the correlation values as an estimation of event correction
        remain_est = calculate_maximum_series_discrepancy(
            labels   = event_remains,
            series_1 = global_corrs,
            series_2 = event_corrs)
    elif method == 'gradient':
        # Calculate the maximum discrepancy in the correlation gradients as an estimation of event correction
        remain_est = calculate_maximum_series_discrepancy(
            labels   = event_remains[1:-1],
            series_1 = [event_corrs[i+1] - event_corrs[i-1] for i in range(1, len(event_corrs)-1)],
            series_2 = [global_corrs[i+1] - global_corrs[i-1] for i in range(1, len(global_corrs)-1)])

    return remain_est





import os, glob

from bamboo.html import png2base64str
from pandda.html import PANDDA_HTML_ENV
from pandda.constants import PanddaAnalyserFilenames, PanddaInspectorFilenames

def write_inspect_html(top_dir, inspector):

    # Get template to be filled in
    template = PANDDA_HTML_ENV.get_template('pandda_summary.html')
    # Output file
    out_file = os.path.join(top_dir, 'results_summaries', PanddaInspectorFilenames.inspect_html)
    # Output directory (for relative symlinks)
    out_dir  = os.path.abspath(os.path.dirname(out_file))

    all_data = inspector.log_table
    len_data = len(all_data.index)

    #####################################################
    #                 Organise Data                     #
    #####################################################

    # Pandda Analysis outputs
    num_blobs = len_data
    num_sites = len(set(all_data['site_idx']))
    num_datasets = len(set(zip(*all_data.index)[0]))

    # Datasets Inspected/Modelled/Empty
    num_fitted = sum(all_data['Ligand Placed'])
    num_viewed = sum(all_data['Viewed'])
    num_empty  = num_viewed - num_fitted
    num_unviewed = len_data - num_viewed

    # Interesting unfitted (bookmarked)
    num_interesting_unfitted = sum(all_data["Interesting"][all_data['Ligand Placed']==False])

    # Number of Confident Ligand Models
    hgh_conf = sum(all_data["Ligand Placed"][all_data['Ligand Confidence']=='High'])
    med_conf = sum(all_data["Ligand Placed"][all_data['Ligand Confidence']=='Medium'])
    low_conf = sum(all_data["Ligand Placed"][all_data['Ligand Confidence']=='Low'])

    # Number of datasets with hits
    try:    num_d_hit = len(set(zip(*all_data.index[all_data['Ligand Placed'] == True])[0]))
    except: num_d_hit = 0
    # Number of sites with hits
    try:    num_s_hit = len(set(all_data[all_data['Ligand Placed'] == True]['site_idx']))
    except: num_s_hit = 0

    # Site Hits
    site_hits = []
    for site_idx in range(1, num_sites+1):
        site_placed = sum(all_data["Ligand Placed"][all_data['site_idx']==site_idx])
        if site_placed > 0: site_hits.append((site_idx, site_placed))

    #####################################################
    #               Prepare html data                   #
    #####################################################

    # ===========================================================>
    # Construct the data object to populate the template
    output_data = {}
    output_data['header'] = 'PANDDA Inspect Summary'
    output_data['title'] = 'PANDDA Inspect Summary'
    output_data['introduction'] = 'Summary of Inspection of Datasets'
    # ===========================================================>
    # Header Images
    output_data['top_images'] = []
    if os.path.exists(os.path.join(top_dir, 'results_summaries', PanddaAnalyserFilenames.pymol_sites_png_1)):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=os.path.join(top_dir, 'results_summaries', PanddaAnalyserFilenames.pymol_sites_png_1))),
                                           'title': 'Identified Sites (Front)' })
    if os.path.exists(os.path.join(top_dir, 'results_summaries', PanddaAnalyserFilenames.pymol_sites_png_2)):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=os.path.join(top_dir, 'results_summaries', PanddaAnalyserFilenames.pymol_sites_png_2))),
                                           'title': 'Identified Sites (Back)' })
    for i_png, png in enumerate(sorted(glob.glob(os.path.join(top_dir, 'results_summaries', PanddaInspectorFilenames.inspect_site_graph_mult).format('*')))):
        output_data['top_images'].append({ 'path': 'data:image/png;base64,{}'.format(png2base64str(path=png)),
                                           'title': 'Identified Site Summary ({})'.format(i_png+1) })
    # ===========================================================>
    # Summary Bar
    output_data['summary_bar'] = []
    output_data['summary_bar'].append({'width':'4', 'colour':'success', 'text':'Datasets w. ligands: {} (of {})'.format(num_d_hit, num_datasets)})
    output_data['summary_bar'].append({'width':'4', 'colour':'success', 'text':'Sites w. ligands: {} (of {})'.format(num_s_hit, num_sites)})
    output_data['summary_bar'].append({'width':'4', 'colour':'warning', 'text':'Marked as interesting (unfitted): {}'.format(num_interesting_unfitted)})

    output_data['summary_bar'].append({'width':'3', 'colour':'info',    'text':'Total number of events: {}'.format(num_blobs)})
    output_data['summary_bar'].append({'width':'3', 'colour':'success', 'text':'High confidence hits:   {}'.format(hgh_conf)})
    output_data['summary_bar'].append({'width':'3', 'colour':'warning', 'text':'Medium confidence hits: {}'.format(med_conf)})
    output_data['summary_bar'].append({'width':'3', 'colour':'danger',  'text':'Low confidence hits:    {}'.format(low_conf)})
    # ===========================================================>
    # Progress Bars
    output_data['progress_bar'] = []
    # Fitting Progress
    output_data['progress_bar'].append({'title':'Fitting Progress', 'data':[]})
    output_data['progress_bar'][-1]['data'].append({'text':'Fitted - {} Events'.format(num_fitted),          'colour':'success', 'size':100.0*num_fitted/len_data})
    output_data['progress_bar'][-1]['data'].append({'text':'Unviewed - {} Events'.format(num_unviewed),      'colour':'warning', 'size':100.0*num_unviewed/len_data})
    output_data['progress_bar'][-1]['data'].append({'text':'No Ligand Fitted - {} Events'.format(num_empty), 'colour':'danger',  'size':100.0*num_empty/len_data})
    # Ligand Distribution
    if num_fitted > 0:
        import math
        weighting = sum([math.log1p(v[1]) for v in site_hits])
        output_data['progress_bar'].append({'title':'Identified Ligands by Site', 'data':[]})
        for i_block, (site_idx, n_hits) in enumerate(sorted(site_hits, key=lambda x: x[1], reverse=True)):
            colour = ('info','default')[i_block%2]
            output_data['progress_bar'][-1]['data'].append({'text':('S{}: {} hit'+'s'*bool(n_hits-1)).format(site_idx, n_hits), 'colour':colour, 'size':100.0*math.log1p(n_hits)/weighting})
    # ===========================================================>
    # Panels
    output_data['top_panels'] = []
    # Site summaries
    for site_idx, site_info in inspector.site_table.iterrows():
        if (site_info['Name'], site_info['Comment']) != ('None','None'):
            output_data['top_panels'].append({  'width':'6','color':'primary',
                                                'title':'<strong>Site {}</strong>: {}'.format(site_idx, site_info['Name']),
                                                'text':'<p><strong>Comment:</strong> '+site_info['Comment']+'</p>'  })
    if num_interesting_unfitted:
        output_data['top_panels'].append({  'width':'12','color':'primary',
                                            'title':'<strong>Interesting Datasets with No Ligand Placed</strong>',
                                            'text':'<p><strong>Datasets:</strong> '+', '.join(['<strong>{}</strong> (Event {})'.format(*i) for i in all_data[(all_data['Ligand Placed']==False)&(all_data['Interesting']==True)].index])+'</p>'  })
    # ===========================================================>
    # Tables
    output_data['table'] = {}
    output_data['table']['column_headings'] = ['Dataset','Viewed','Interesting','Lig. Placed','Event','Site','1 - BDC','Z-Peak','Map Res.','Map Unc.','Confidence','Comment','']
    output_data['table']['rows'] = []
    # Add the datasets as rows
    for i_d in range(len(all_data.index)):

        # colour choices - 'success', 'muted', 'danger'
        # icon choices   - 'ok', 'flag', 'remove'

        d_data = all_data.iloc[i_d].to_dict()
        d_tag, d_event = all_data.index[i_d]

        columns = []

        if d_data['Viewed']:        columns.append({'colour':'success', 'icon':'ok',     'message':d_data['Viewed']})
        else:                       columns.append({'colour':'danger',  'icon':'remove', 'message':d_data['Viewed']})

        if d_data['Interesting']:   columns.append({'colour':'success', 'icon':'ok',     'message':d_data['Interesting']})
        else:                       columns.append({'colour':'danger',  'icon':'remove', 'message':d_data['Interesting']})

        if d_data['Ligand Placed']: columns.append({'colour':'success', 'icon':'ok',     'message':d_data['Ligand Placed']})
        else:                       columns.append({'colour':'danger',  'icon':'remove', 'message':d_data['Ligand Placed']})

        columns.append({'message':d_event})
        columns.append({'message':d_data['site_idx']})
        columns.append({'message':round(d_data['1-BDC'],3)})
        columns.append({'message':round(d_data['z_peak'],3)})
        columns.append({'message':d_data['analysed_resolution']})
        columns.append({'message':round(d_data['map_uncertainty'],3)})

        columns.append({'message':d_data['Ligand Confidence']})
        columns.append({'message':d_data['Comment']})

        row_message = 'Hit' if d_data['Ligand Placed'] else \
                      ''
        row_colour  = 'success' if d_data['Ligand Placed'] else \
                      'danger' if not d_data['Viewed'] else \
                      'info'

        output_data['table']['rows'].append({'heading' : d_tag,
                                             'colour'  : row_colour,
                                             'message' : row_message,
                                             'columns' : columns})

    with open(out_file, 'w') as out_html:
        out_html.write(template.render(output_data))



import os,sys

from bamboo.common.command import CommandManager

def fft_mtz_to_map(mtz_file, map_file, cols):
    """Converts an MTZ Format File to a MAP File (using fft as default)"""

    # Initialise
    writer = CommandManager('fft')
    # Set Program Arguments
    writer.add_command_line_arguments('hklin',mtz_file,'mapout',map_file)
    # Set Program Input
    writer.add_standard_input(['LABIN F1={!s} PHI={!s}'.format(cols['F'],cols['P']),'END'])
    # RUN!
    writer.run()
    # Check Output
    if writer.process.returncode != 0:
        print('\nOUT\n\n'+writer.out)
        print('\nERR\n\n'+writer.err)
        raise RuntimeError('fft failed to generate map from {!s}'.format(mtz_file))

    return writer

def mask_map(mapin, maskpdb, mapout, border=1):
    """Takes mapin and masks around atoms in maskpdb"""

    # Masking object
    masker = CommandManager('mapmask')
    # Set input files
    masker.add_command_line_arguments(['mapin',mapin,'mapout',mapout,'xyzin',maskpdb])
    # Set stdin
    masker.add_standard_input(['BORDER {!s}'.format(border),'END'])
    # Run!
    masker.run()
    # Report errors
    if masker.process.returncode!=0:
        raise RuntimeError('mapmask failed to mask map {!s}'.format(mapin))

    # Return Command Managers for flexible handling of out & err
    return masker

def create_asu_map(mapin, mapout):
    """Takes mapin and masks to the asymmetric unit"""

    # Masking object
    masker = CommandManager('mapmask')
    # Set input files
    masker.add_command_line_arguments(['mapin',mapin,'mapout',mapout])
    # Set stdin
    masker.add_standard_input(['XYZLIM ASU','END'])
    # Run!
    masker.run()
    # Report errors
    if masker.process.returncode!=0:
        raise RuntimeError('mapmask failed to create asu map from {!s}'.format(mapin))

    # Return Command Managers for flexible handling of out & err
    return masker




from rdkit import Chem

from bamboo.constants import WATER_NAMES
from bamboo.rdkit_utils.mol import check_pdb_readable

def order_structures_by_minimum_distance_to_reference(refpdb, pdbs, mincutoff=1.5, maxcutoff=6):
    """Return the molecules have at least one non-H atom within a certain distance of the reference structure"""

    # Deal with cases where mincutoff, maxcutoff are not specified
    if not mincutoff:
        mincutoff = 0
    if not maxcutoff:
        maxcutoff = 9999999

    pdbs_to_return = []

    refmol = check_pdb_readable(refpdb)

    for pdbfile in pdbs:
        pdbmol = check_pdb_readable(pdbfile)

        min_dist = calculate_minimum_distance_between_mols(refmol, pdbmol)

        # Reject if there is a clash
        if min_dist >= mincutoff and min_dist <= maxcutoff:
            pdbs_to_return.append((min_dist,pdbfile))

    # Order by minimum distance
    sorted_files = sorted(pdbs_to_return, key=lambda tup: tup[0])

    return [t[1] for t in sorted_files]

def order_structures_by_number_of_contacts_to_reference(refpdb, pdbs, cutoff=3.5):
    """Calculate the number of contacts between `pdb` and refpdb, and return a list of decreasing contacts"""

    refmol = check_pdb_readable(refpdb)

    contacts_list = []

    for pdbfile in pdbs:
        pdbmol = check_pdb_readable(pdbfile)

        # Add the number of contacts and the pdbfile to the list
        contacts_list.append((calculate_number_of_pairwise_distances_within_cutoff(refmol, pdbmol, cutoff=cutoff), pdbfile))

    return sorted(contacts_list, key=lambda tup: tup[0], reverse=True)

def calculate_minimum_distance_between_mols(mol1, mol2):
    """Takes two molecules and returns the minimum distance between the two structures"""

    distances = calculate_pairwise_distances_between_mols(mol1, mol2)

    if not distances:
        return None
    else:
        return min(distances)

def calculate_number_of_pairwise_distances_within_cutoff(mol1, mol2, cutoff=3.5):
    """Count number of times a distance less than `cutoff` is observed between an atom of mol1 and an atom of mol2"""

    distances = calculate_pairwise_distances_between_mols(mol1, mol2)

    if not distances:
        return None

    within_cutoff = [(d<cutoff) for d in distances]
    counts = within_cutoff.count(True)

    return counts

def calculate_pairwise_distances_between_mols(mol1, mol2):
    """Calculates pairwise distances between atoms in mol1 and mol2"""

    # TODO Change this so that it runs through all of the conformations...
    conf1 = mol1.GetConformer()
    conf2 = mol2.GetConformer()

    # Get the indexes of the atoms that aren't water
    mol1_idxs = [a.GetIdx() for a in mol1.GetAtoms() if a.GetMonomerInfo().GetResidueName() not in WATER_NAMES]
    mol2_idxs = [a.GetIdx() for a in mol2.GetAtoms() if a.GetMonomerInfo().GetResidueName() not in WATER_NAMES]

    # Iterate through all pairs of atoms and calculate pairwise distances
    distances = [conf1.GetAtomPosition(i1).Distance(conf2.GetAtomPosition(i2)) for i2 in mol2_idxs for i1 in mol1_idxs]

    return distances



from giant.xray.miller.utils import apply_symmetry_to_miller_array, scale_amplitude_arrays_via_intensities

def apply_simple_scaling(miller, ref_miller):
    """Applies a simple scaling algorithm to two similar, but with non-identical unit cells, datasets"""
    miller_sym       = apply_symmetry_to_miller_array(chg_miller=miller, ref_miller=ref_miller)
    miller_sym_scale = scale_amplitude_arrays_via_intensities(chg_arr_a=miller_sym, ref_arr_a=ref_miller, n_bins=20)
    miller_scale     = apply_symmetry_to_miller_array(chg_miller=miller_sym_scale, ref_miller=miller)

    return miller_scale

def divide_miller_indices_into_equal_bins(miller, indices_per_bin):
    """Analyse miller indices and split into equally populated bins"""

    pass

def scale_and_truncate_miller_array(miller, ref_miller, low_index_cutoff, high_index_cutoff, index_bins=None):
    """Apply scaling in resolution bins. Miller arrays are then truncated at low and high miller indices. Scaling bins can be predetermined."""

    pass


if __name__ == '__main__':

    import iotbx.mtz

    m = iotbx.mtz.object('./reference.mtz')

#    c = m.extract_complex('FWT', 'PHWT')
    a=m.as_miller_arrays_dict()[('XDScrystal', 'XDSdataset', 'FWT')]



import math

from bamboo.rdkit_utils.mol import check_pdb_readable
from bamboo.rdkit_utils.bonds.dihedral import calculate_dihedral_angle_differences

def calculate_dihedral_angle_rmsd_from_file(file1, file2):
    """Calculate angle rmsd between structures"""
    return calculate_dihedral_angle_rmsd(check_pdb_readable(file1),check_pdb_readable(file2))

def calculate_dihedral_angle_rmsd(mol1, mol2):
    """Calculate angle rmsd between structures"""

    try:
        # Get the atoms and the differences
        atoms, diffs = calculate_dihedral_angle_differences(mol1, mol2)
    except EqualityError as Err:
        print Err
        return None

    # List of rmsds for different ways of matching
    rmsds = []
    # Iterate through the different ways of matching the atoms
    for diff in diffs:
        if not diff:
            continue
        else:
            sqdiff = [d**2 for d in diff]
            # Calculate the RMSD
            rmsds.append(math.sqrt(sum(sqdiff)/len(sqdiff)))
    # Return the minimum rmsd
    if rmsds:
        return min(rmsds)
    else:
        return None




import os, sys, math

from rdkit import Chem
from rdkit.Chem import BRICS
from bamboo.macro.molecule import MacroMol
from bamboo.rdkit_utils.mol import check_pdb_readable

def break_on_rotatable_bonds_to_mol(inmol):
    """Takes a mol and breaks it on all of the rotatable bonds (well, most of them) - returns a single mol, or None if there are no rotatable bonds"""

    # Get the indices of the atoms around the bonds to be broken
    atom_pairs = [p[0] for p in BRICS.FindBRICSBonds(inmol) if p]
    # Return if no bonds found... as it was given
    if not atom_pairs:
        return inmol
    # Get the bond indices
    bonds = [inmol.GetBondBetweenAtoms(at1,at2).GetIdx() for (at1, at2) in atom_pairs]
    # Fragment the molecule
    fragged_mol = Chem.rdmolops.FragmentOnBonds(inmol, bonds, addDummies=False)

    return fragged_mol

def break_on_rotatable_bonds_to_mols(inmol):
    """Takes a mol and breaks it on all of the rotatable bonds (well, most of them) - returns a set of mols: one for each fragment, or [] if there are no rotatable bonds"""

    fragged_mol = break_on_rotatable_bonds_to_mol(inmol)
    # No fragged mol => No bonds to fragment => return original mol
    if not fragged_mol:
        raise RDkitFragmentationError('Failed to break on rotatable bonds')
    # Get the individual mols for the fragments
    frags = Chem.GetMolFrags(fragged_mol,asMols=True)

    return frags

def break_and_rename_mol_to_file(infile, outfile):
    """Takes a ligand, breaks it up and renames the different fragments with inscodes. compiles into one file"""

    inmol = check_pdb_readable(infile)
    # PDB-rdkit disagreement...
    if not inmol:
        raise RDkitReadError('Cannot read {!s}'.format(infile))

    orig_mol = MacroMol(infile)

    assert len(orig_mol.getChains())==1, 'MORE THAN ONE CHAIN PRESENT IN FILE! {!s}'.format(infile)
    assert len(orig_mol.getResidues())==1, 'MORE THAN ONE RESIDUE PRESENT IN FILE! {!s}'.format(infile)

    # Keep the atom numbers (these are conserved in the mappings)
    atom_numbers = dict([(a.atomname,a.serialnum) for a in orig_mol.getResidues()[0].atoms])

    # Break the molecule into fragments (or get inmol back if no rotatable bonds)
    frags = break_on_rotatable_bonds_to_mols(inmol)

    # List of output pdb blocks
    out_blocks = []

    for i, frag in enumerate(frags):

        raw_block = Chem.MolToPDBBlock(frag)
        raw_frag = MacroMol(raw_block)
        # Check there's only one residue in file
        assert len(raw_frag.getChains())==1, 'FRAGMENT HAS MORE THAN ONE CHAIN? {!s}\n{!s}'.format(infile, raw_block)
        assert len(raw_frag.getResidues())==1, 'FRAGMENT HAS MORE THAN ONE RESIDUE? {!s}\n{!s}'.format(infile, raw_block)
        # Form the new inscode
        new_inscode = chr(i + ord('A'))
        # Change the inscode
        residue = raw_frag.getResidues()[0]
        residue.update_inscode(new_inscode)
        # Change the atom numbers back to those in the original file for continuity
        residue.update_atom_numbers(atom_numbers)
        # Get the pdb string and use
        proc_block = residue.getPdbString()
        # Check for consistency and append
        out_blocks.append(proc_block)

    # Add easy-header
    header_block = ''.join(orig_mol.easy_header).strip('\n')
    # Add easy-footer
    footer_block = ''.join(orig_mol.easy_footer).strip('\n')
    # Create atom block
    out_block = ''.join(out_blocks).replace('END','').strip('\n')
    # Join the blocks together
    block_to_write = '\n'.join([header_block, out_block, footer_block, 'END\n'])
    # Check for consistency
    block_to_write = block_to_write.replace('\n\n','\n')

    open(outfile,'w').write(block_to_write)

    return outfile





from scipy import stats as scipy_stats
import numpy

class data_collection(object):
    def __init__(self):
        """Stores observations across a set of datasets"""
        self.data = {}
        self.entry_ids = None
        self._data_length = None

    def set_data_length(self, data_length):
        """Set the mask length"""
        if self._data_length: assert self._data_length == data_length, "CAN'T CHANGE THE DATASET LENGTH"
        else:                 self._data_length = data_length

    def set_entry_ids(self, entry_ids):
        """List of ids that label the different observations"""
        assert not self.entry_ids, 'DATASET IDS HAVE ALREADY BEEN SET'
        if self._data_length: assert len(entry_ids) == self._data_length
        else:                 self._data_length = len(entry_ids)
        self.entry_ids = entry_ids

    def add_empty_data(self, data_name):
        """Add a new empty set of data"""
        if data_name in self.data.keys():
            print('PROGRAM WARNING: OVERWRITING DATA IN DATA SUMMARY: {!s}'.format(data_name))
        self.data[data_name] = [numpy.nan]*self._data_length

    def add_data(self, data_name, data_values):
        """Add a new set of data"""
        if data_name in self.data.keys():
            print('PROGRAM WARNING: OVERWRITING DATA IN DATA SUMMARY: {!s}'.format(data_name))
        assert len(data_values) == self._data_length
        self.data[data_name] = data_values

    def get_data(self, data_name):
        """Get a single set of data"""
        return self.data[data_name]

    def has_data(self, data_name):
        """Check if it has a single mask"""
        return data_name in self.data.keys()

    def set_data_value(self, data_name, entry_id, value):
        """Set a particular entry in the data list `data_name`, corresponding to `entry_id` to `value`"""
        assert data_name in self.data.keys()
        assert entry_id in self.entry_ids
        self.data[data_name][self.entry_ids.index(entry_id)] = value

    def get_data_value(self, data_name, entry_id):
        """Retrieve an observation for a particular entry_id"""
        assert data_name in self.data.keys()
        assert entry_id in self.entry_ids
        return self.data[data_name][self.entry_ids.index(entry_id)]

    def get_data_as_z_values(self, data_name):
        """Convert a set of data to z-values"""
        data_values = self.get_data(data_name=data_name)
        z_values = scipy_stats.zscore(a=data_values).tolist()
        return zip(self.entry_ids, z_values)

    def get_data_as_dict(self, data_name):
        """Get a single set of data - returned as a dictionary with the entry_ids as keys"""
        return dict(zip(self.entry_ids, self.get_data(data_name=data_name)))

class multiple_data_collection(object):
    def __init__(self):
        """Stores multiple related data_collection objects - Allows the same observation to be pulled from all of the data_collections"""

        # Contains keys for the dictionaries
        self.collection_ids = []
        # Contains the collection objects
        self.collections = {}
        # Contains information about each collection object
        self.collection_info = {}
        # Number of collection objects
        self._collection_length = None

    def set_collection_ids(self, collection_ids):
        """Set the ids of the different collections"""
        if self._collection_length: assert len(collection_ids) == self._collection_length
        else:                       self._collection_length = len(collection_ids)
        assert not self.collection_ids, 'COLLECTION IDS HAVE ALREADY BEEN SET'
        self.collection_ids = collection_ids

    def add_collection(self, collection_id, collection):
        """Add a data collection"""
        assert collection_id in self.collection_ids, 'INVALID COLLECTION ID'
        self.collections[collection_id] = collection

    def get_collection(self, collection_id):
        """Get a data collection"""
        return self.collections[collection_id]

    def add_collection_info(self, info_name, info_values):
        """Store info about the different data collections"""
        assert self._collection_length == len(info_values)
        self.collection_info[info_name] = dict(zip(self.collection_ids, info_values))

    def get_collection_info(self, info_name):
        """Get info about the different data collections"""
        return self.collection_info[info_name]

    def get_data_from_collections(self, data_name):
        """Retrieve a zipped list of collection ids, and the data contained within them referenced by data_name"""
        return [(col_id, self.get_collection(col_id).get_data(data_name)) for col_id in self.collection_ids]




import iotbx.pdb
from scitbx.math import dihedral_angle

from giant.structure.select import extract_backbone_atoms
from giant.structure.iterators import generate_residue_triplets

########################################################################################

def get_all_phi_psi_for_hierarchy(pdb_h, deg=True):
    """Calculate all phi-psi angles for a structure"""
    for prev, current, next in generate_residue_triplets(pdb_h):
        phi, psi = calculate_residue_phi_psi_angles(prev=prev, current=current, next=next, deg=deg)
        yield (current, (phi,psi))

########################################################################################

def calculate_residue_phi_psi_angles(prev, current, next, deg=True):
    """Calculate phi-psi angles for a set of three residues. Returns (phi, psi)."""
    phi = dihedral_angle(sites=extract_phi_sites(prev=prev, current=current), deg=deg)
    psi = dihedral_angle(sites=extract_psi_sites(current=current, next=next), deg=deg)
    return (phi, psi)

########################################################################################

def extract_psi_sites(current, next):
    """Extract sites to calculate psi angle from the selected residues"""
    return [a.xyz for a in extract_psi_atoms(current=current, next=next)]

def extract_phi_sites(prev, current):
    """Extract sites to calculate phi angle from the selected residues"""
    return [a.xyz for a in extract_phi_atoms(prev=prev, current=current)]

########################################################################################

def extract_psi_atoms(current, next):
    """Extract atoms to calculate phi angle from the selected residues"""
    n_curr = ca_curr = c_curr = n_next = None
    n_next, ca_next, c_next = extract_backbone_atoms(residue=next)
    n_curr, ca_curr, c_curr = extract_backbone_atoms(residue=current)
    return (n_curr, ca_curr, c_curr, n_next)

def extract_phi_atoms(prev, current):
    """Extract sites to calculate phi angle from the selected residues"""
    c_prev = n_curr = ca_curr = c_curr = None
    n_prev, ca_prev, c_prev = extract_backbone_atoms(residue=prev)
    n_curr, ca_curr, c_curr = extract_backbone_atoms(residue=current)
    return (c_prev, n_curr, ca_curr, c_curr)

########################################################################################





from scitbx.array_family import flex

def extract_ligands_from_hierarchy(pdb_hierarchy):
    """Extracts ligands from hierarchy objects. Returns as new hierarchy objects"""

    # Filter out waters
    no_waters = pdb_hierarchy.select(pdb_hierarchy.atom_selection_cache().selection('not water'))
    # Get only hetatms
    hetero_only = no_waters.select(flex.bool([a.hetero for a in no_waters.atoms_with_labels()]))
    hetero_cache = hetero_only.atom_selection_cache()

    # Build chain-resnum combinations
    res_ids = []
    for ch in hetero_only.chains():
        for res in ch.residue_groups():
            res_ids.append((ch.id, res.resseq, res.icode))

    roots = []
    # Select each residue at a time
    for chn, seq, icode in res_ids:
        atm_sel = hetero_cache.selection("chain '{!s}' and resseq '{!s}' and icode '{!s}'".format(chn,seq,icode))
        new_root = hetero_only.select(atm_sel)
        roots.append(new_root)

    return roots



import itertools
import scipy.spatial
import scipy.cluster
import numpy

def generate_group_idxs(group_vals):
    num_vals = len(group_vals)
    # Sort the list idxs by group val
    sorted_idx = sorted(range(num_vals), key=lambda i: group_vals[i])
    # Reorder the group vals
    sorted_val = [group_vals[i] for i in sorted_idx]
    # For each group val, return idxs for this group
    for val, sel_idxs in itertools.groupby(range(num_vals), key=lambda i: sorted_val[i]):
        yield val, [sorted_idx[idx] for idx in sel_idxs]

def find_connected_groups(connection_matrix):
    """Take a matrix of connected elements and output groups"""

    assert not set(connection_matrix.flatten()).difference(set([0,1])), 'CONNECTION MATRIX MUST CONSIST OF 1s AND 0s ONLY'
    assert connection_matrix.ndim == 2, 'CONNECTION MATRIX MUST BE OF DIMENSION 2'
    assert connection_matrix.shape[0] > 1, 'MATRIX MUST BE LARGER THAN 1x1'
    assert connection_matrix.shape[0] == connection_matrix.shape[1], 'CONNECTION MATRIX MUST BE SQUARE'

    # Make it symmetrical - if 1 is connected to 2, 2 is connected to 1
    sym_connection_matrix = ((connection_matrix + connection_matrix.T) != 0).astype(int)
    # Convert from a connection array to a distance array (0 for connected, 1 for unconnected)
    dist_mx = 1 - sym_connection_matrix
    # Convert to condensed distance matrix
    cd_dist_mx = scipy.spatial.distance.squareform(dist_mx)
    # Calculate the linkage matrix
    l = scipy.cluster.hierarchy.linkage(cd_dist_mx, method='single', metric='euclidean')
    # Cluster with very low cutoff
    connected_groups = scipy.cluster.hierarchy.fcluster(Z=l, t=0.01, criterion='distance').tolist()

    return connected_groups




import os, sys, copy, re
import time

import libtbx.phil

import numpy

import iotbx.pdb
import scipy.spatial

from scitbx.array_family import flex
from giant.structure.align import perform_flexible_alignment, find_nearest_calphas, transform_coordinates_with_flexible_alignment

#######################################

blank_arg_prepend = 'pdb='

master_phil = libtbx.phil.parse("""
input {
    reference_pdb = None
        .type = path
    pdb = None
        .type = path
        .multiple = True
}
alignment {
    cutoff_radius = 10.0
        .type = float
}
output {
    prefix = aligned_
        .type = path
}
""")

#######################################

def run(params):

    assert params.input.reference_pdb is not None, 'No reference_pdb provided'

    # =================================================>
    # Load structure
    # =================================================>
    ref_hierarchy = iotbx.pdb.hierarchy.input(params.input.reference_pdb).hierarchy

    t_start = time.time()

    for mov_pdb in params.input.pdb:

        print 'ALIGNING: {}'.format(mov_pdb)

        # =================================================>
        # Load structure
        # =================================================>
        mov_hierarchy = iotbx.pdb.hierarchy.input(mov_pdb).hierarchy
#        mov_hierarchy_trimmed = mov_hierarchy.select(mov_hierarchy.atom_selection_cache().selection('pepnames and (name CA or name C or name O or name N)'))

        # =================================================>
        # Calculate alignments
        # =================================================>
        alignments = perform_flexible_alignment( mov_hierarchy = ref_hierarchy,
                                                 ref_hierarchy = mov_hierarchy,
                                                 cutoff_radius = params.alignment.cutoff_radius    )

        # =================================================>
        # Find which alignments should be used for coords
        # =================================================>
        mappings = find_nearest_calphas( hierarchy   = mov_hierarchy,
                                         coordinates = mov_hierarchy.atoms().extract_xyz()   )

        # =================================================>
        # Transform coordinates
        # =================================================>
        aligned_coords = transform_coordinates_with_flexible_alignment( alignments  = alignments,
                                                                        coordinates = mov_hierarchy.atoms().extract_xyz(),
                                                                        mappings    = mappings,
                                                                        inverse     = True   )

        # =================================================>
        # Output structure
        # =================================================>
        # Copy the input structure
        new_hierarchy = mov_hierarchy.deep_copy()
        new_hierarchy.atoms().set_xyz(aligned_coords)
        new_hierarchy.write_pdb_file(file_name=os.path.join(os.path.dirname(mov_pdb), params.output.prefix+os.path.basename(mov_pdb)))

    t_end = time.time()
    print 'Total Runtime: {!s}'.format(time.strftime("%H hours:%M minutes:%S seconds", time.gmtime(t_end-t_start)))

#######################################

if __name__ == '__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



from bamboo.macro.molecule import MacroMol, MacroMolError
from bamboo.rdkit_utils.smile import find_structure_matches

def get_residue_labels(pdb_file):
    """Takes the PDB file and returns a list of the residues labels in the form (res.resname, res.chain, res.resnum, res.inscode) (e.g. for edstats)"""

    labels = []
    # Create a structure object
    structure = MacroMol(pdb_file)
    # Iterate through and create labels
    for res in structure.get_residues():
        labels.append(res.get_res_id())

    return labels

def get_mean_occupancy(pdb_file, type='aminos'):
    """Gets the mean occupancy of the structure in pdb_file. Changing `type` changes the residues that are used to calculate the occupancy. Choices are ['all','aminos','ions','waters']"""

    m = MacroMol(pdb_file)

    occupancies = []

    if type=='all':
        residues = m.get_residues()
    elif type=='aminos':
        residues = m.get_aminos()
    elif type=='waters':
        residues = m.get_waters()
    elif type=='ions':
        residues = m.get_ions()

    [occupancies.append(res.get_mean_occupancy()) for res in residues]

    mean_occ = sum(occupancies)/len(occupancies)

    return mean_occ

def get_residue_occupancies(pdb_file):
    """Gets the occupancies of individual residues. Returns a dict"""

    m = MacroMol(pdb_file)

    occupancies = {}

    for res in m.get_residues():
        occupancies[res.get_res_id()] = res.get_mean_occupancy()

    return occupancies

def extract_compound_smiles(pdbin, find_boring_compounds=False):
    """Takes a PDB file and returns smiles strings of all of the compounds in the file"""

    try:
        m = MacroMol(pdbin)
    except MacroMolError:
        raise

    cpds = m.get_unknowns()

    # ADD FILTERING STEP!
    if find_boring_compounds:
        cpds.extend(m.get_solvent())

    [c.get_smiles() for c in cpds]
    return [c.smile for c in cpds if c.smile]




import os, sys, glob

import libtbx.phil

from bamboo.common.command import CommandManager
from bamboo.common.path import rel_symlink

#######################################

blank_arg_prepend = {   '.mtz':'mtz='   }

master_phil = libtbx.phil.parse("""
input {
    mtz = 'refine.mtz'
        .multiple = True
        .type = str
}
settings {
    f_obs = F,SIGF
        .type = str
    f_calc = F-model,PHIF-model
        .type = str
}

""")

#######################################

def run(params):

    assert params.input.mtz, 'No MTZs given for comparison'

    assert not os.path.exists('merged.mtz'), 'A file called merged.mtz already exists. Please delete it before re-running.'

    merge = CommandManager('giant.mtz.merge')
    merge.add_command_line_arguments( 'label_suffix=incremental' )
    merge.add_command_line_arguments( params.input.mtz )
    merge.add_command_line_arguments( params.settings.f_obs.split(',') )
    merge.add_command_line_arguments( params.settings.f_calc.split(',') )
    merge.run()
    merge.write_output('merged.log')

    assert os.path.exists('merged.mtz')

    fo1, fo2 = params.settings.f_obs.split(',')
    fc1, fc2 = params.settings.f_calc.split(',')

    for i_2 in range(1, len(params.input.mtz)+1):
        for i_1 in range(1, len(params.input.mtz)+1):
            if i_1 == i_2:
                break
            match = CommandManager('cphasematch')
            match.add_command_line_arguments( ['-mtzin', 'merged.mtz'] )
            match.add_command_line_arguments( ['-mtzout', 'phasematch.mtz'] )
            match.add_command_line_arguments( ['-colin-fo',   '{}-{},{}-{}'.format(fo1,i_1,fo1,i_1)] )
            match.add_command_line_arguments( ['-colin-fc-1', '{}-{},{}-{}'.format(fc1,i_1,fc1,i_1)] )
            match.add_command_line_arguments( ['-colin-fc-2', '{}-{},{}-{}'.format(fc2,i_2,fc2,i_2)] )
            match.run()
            march.write_output('phasematch-{}-{}.log'.format(i_1.i_2))


#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



# Lists/Constants

# =================== #
# Refiner Settings    #
# =================== #

refiners = ['refmac','phenix.refine']
allowed_refiner_args = []

# =================== #
# Ligand Generation   #
# =================== #

ligand_builders = ['elbow','grade','writedict']
#ligand_builders = ['elbow','grade','prodrg','writedict']
allowed_builder_args = []

# =================== #
# Ligand Fitting      #
# =================== #

ligand_fitters = ['ligandfit','rhofit','flynn']
#ligand_fitters = ['phenix.ligandfit','arpwarp','rhofit','coot','jligand','flynn','afitt']
allowed_fitter_args = []

# =================== #

def get_ligand_builder(program):
    """Get Ligand Builder Object for `program`"""

    from bamboo.wrappers.ligand_builders.elbow import ElbowObject
    from bamboo.wrappers.ligand_builders.grade import GradeObject
    from bamboo.wrappers.ligand_builders.prodrg import ProdrgObject
    from bamboo.wrappers.ligand_builders.writedict import WritedictObject

    # Check it's a valid program
    if program not in ligand_builders:      raise LigandBuilderSelectionError('{!s} not in {!s}'.format(program, ', '.join(ligand_builders)))

    if program in ['elbow','phenix.elbow']:     return ElbowObject()
    elif program == 'grade':                    return GradeObject()
    elif program == 'prodrg':                   return ProdrgObject()
    elif program in ['writedict','afitt']:      return WritedictObject()
    else:
        raise LigandBuilderSelectionError('MY BAD. The Code for {!s} has not been written yet...'.format(program))

def get_ligand_fitter(program):
    """Get Ligand Fitter Object for `program`"""

    from bamboo.wrappers.ligand_fitters.ligandfit import LigandfitObject
    from bamboo.wrappers.ligand_fitters.arpwarp import ArpwarpObject
    from bamboo.wrappers.ligand_fitters.rhofit import RhofitObject
    from bamboo.wrappers.ligand_fitters.flynn import FlynnObject
    from bamboo.wrappers.ligand_fitters.coot import CootObject

    # Check it's a valid program
    if program not in ligand_fitters:       raise LigandFitterSelectionError('{!s} not in {!s}'.format(program, ', '.join(ligand_fitters)))

    if program in ['ligandfit','phenix.ligandfit']:     return LigandfitObject()
    elif program == 'arpwarp':                          return ArpwarpObject()
    elif program == 'rhofit':                           return RhofitObject()
    elif program in ['flynn','afitt']:                  return FlynnObject()
    elif program == 'coot':                             return CootObject()
    else:
        raise LigandFitterSelectionError('MY BAD. The Code for {!s} has not been written yet...'.format(program))

def get_refiner(program):
    '''Returns Initialised Refiner Object'''

    from bamboo.wrappers.refiners.phenix import PhenixrefineObject
    from bamboo.wrappers.refiners.refmac import RefmacObject

    # Check it's a valid program
    if program not in refiners:
        raise RefinerSelectionError('{!s} not in {!s}'.format(program, ', '.join(refiners)))

    if program.lower()=='phenix.refine':                return PhenixrefineObject()
    elif program.lower()=='refmac':                     return RefmacObject()
    else:
        raise RefinerSelectionError('MY BAD. The Code for {!s} has not been written yet...'.format(program))




import os, sys

import libtbx.phil
from bamboo.common.command import CommandManager
from giant.xray.data import CrystalSummary

#######################################

blank_arg_prepend = 'mtz='

master_phil = libtbx.phil.parse("""
input {
    mtz = None
        .type = path
        .multiple = True
    columns = F,SIGF
        .type = str
}
method {
    package = *phenix ccp4
        .type = choice
}
output {
    column = IMEAN
        .type = str
    file_prefix = intensities_
        .type = str
}
""")

#######################################

def run(params):

    input_cols = params.input.columns.split(',')

    for mtz_file in params.input.mtz:

        output_mtz = os.path.join(os.path.dirname(mtz_file), params.output.file_prefix+os.path.basename(mtz_file))

        cs = CrystalSummary.from_mtz(mtz_file=mtz_file)

        if params.output.column in cs.column_labels:
            print 'OUTPUT COLUMN ALREADY PRESENT IN MTZ_FILE: {}'.format(mtz_file)
            if not os.path.exists(output_mtz):
                os.symlink(os.path.basename(mtz_file), output_mtz)
            continue

        cols_correct  = True
        for col in input_cols:
            if col not in cs.column_labels:
                cols_correct = False
                break
        if cols_correct == False:
            print 'Column not present in mtz_file:'
            print cs.column_labels
            print 'SKIPPING'
            continue

        if params.method.package == 'phenix':
            cm = CommandManager('phenix.reflection_file_converter')
            cm.add_command_line_arguments([ mtz_file,
                                            '--mtz-root-label='+params.output.column,
                                            '--label='+params.input.columns,
                                            '--non-anomalous',
                                            '--write-mtz-intensities',
                                            '--mtz='+output_mtz     ])
        elif params.method.package == 'ccp4':
            raise Exception('NOT IMPLEMENTED')

        cm.print_settings()
        cm.run()
        print '============================>'
        print cm.output
        print '============================>'
        print cm.error
        print '============================>'

#######################################

if __name__=='__main__':
    from giant.jiffies import run_default
    run_default(run=run, master_phil=master_phil, args=sys.argv[1:], blank_arg_prepend=blank_arg_prepend)



import os, time

from bamboo.wrappers.ligand_fitters import FitterObject

class ArpwarpObject(FitterObject):

    def _custom_init(self):
        self.program = 'auto_ligand.sh'
        self.name = 'arpwarp'

    def _create_program_arguments(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):
        """Run ARPwARP to fit the ligand"""

        # Run the custom parts of this function
        self.outscores = self.outtemplate+'.scores'
# FIXME        (F,P,Rf) = GetMTZRefinementHeadings(mtz)

        # Format Settings
        cmd_line_flags, std_inpt_flags = self._convert_flags(flags)
        # Form Command Line Args
        cmd_line_args = ['datafile',mtz,'protein',apopdb,'ligand',ligpdb,'workdir',outdir,'extralibrary',ligcif,'fp',F,'sigfp',P,'freer',Rf,'ligandfileout',self.outpdb]
        cmd_line_args.extend(cmd_line_flags)
        # Form Standard Input
        std_inpt_args = []
        std_inpt_args.extend(std_inpt_flags)

        return cmd_line_args, std_inpt_args

    def _post_process(self, ligcif, ligpdb, mtz, apopdb, outdir, outfile, flags):

        # Get the output models
        self._find_output_files()

        return

    def _convert_flags(self, flags):
        """Take the flags passed and generate the command string appropriately"""

        cmd_line_flags = []
        std_inpt_flags = []

        for flag in flags:
            pass

        return cmd_line_flags, std_inpt_flags

    def _find_output_files(self):
        """Reads the ARPwARP summary to find the output files"""
        pass

        return

